"""
Distributional optimal power flow (DOPF).
"""
import numpy as np

from ams.core.param import RParam

from ams.routines.dcopf import DCOPF

from ams.opt.omodel import Var, Constraint, Objective


class DOPF(DCOPF):
    """
    Linearzied distribution OPF, where power loss are ignored.

    UNDER DEVELOPMENT!

    Reference:

    [1] L. Bai, J. Wang, C. Wang, C. Chen, and F. Li, “Distribution Locational Marginal Pricing (DLMP)
    for Congestion Management and Voltage Support,” IEEE Trans. Power Syst., vol. 33, no. 4,
    pp. 4061–4073, Jul. 2018, doi: 10.1109/TPWRS.2017.2767632.
    """

    def __init__(self, system, config):
        DCOPF.__init__(self, system, config)
        self.info = 'Linearzied distribution OPF'
        self.type = 'DED'

        # -- Data Section --
        # ---  generator ---
        self.qmax = RParam(info='generator maximum reactive power',
                           name='qmax', tex_name=r'q_{max}', unit='p.u.',
                           model='StaticGen', src='qmax',)
        self.qmin = RParam(info='generator minimum reactive power',
                           name='qmin', tex_name=r'q_{min}', unit='p.u.',
                           model='StaticGen', src='qmin',)
        # --- load ---
        self.qd = RParam(info='reactive demand',
                         name='qd', tex_name=r'q_{d}', unit='p.u.',
                         model='StaticLoad', src='q0',)
        # --- bus ---
        self.vmax = RParam(info="Bus voltage upper limit",
                           name='vmax', tex_name=r'v_{max}',
                           unit='p.u.',
                           model='Bus', src='vmax', no_parse=True,
                           )
        self.vmin = RParam(info="Bus voltage lower limit",
                           name='vmin', tex_name=r'v_{min}',
                           unit='p.u.',
                           model='Bus', src='vmin', no_parse=True,)
        # --- line ---
        self.r = RParam(info='line resistance',
                        name='r', tex_name='r', unit='p.u.',
                        model='Line', src='r',
                        no_parse=True,)
        self.x = RParam(info='line reactance',
                        name='x', tex_name='x', unit='p.u.',
                        model='Line', src='x',
                        no_parse=True,)
        # --- Model Section ---
        # --- generator ---
        self.qg = Var(info='Gen reactive power',
                      name='qg', tex_name=r'q_{g}', unit='p.u.',
                      model='StaticGen', src='q',)
        self.qglb = Constraint(name='qglb', is_eq=False,
                               info='qg min',
                               e_str='-qg + mul(ug, qmin)',)
        self.qgub = Constraint(name='qgub', is_eq=False,
                               info='qg max',
                               e_str='qg - mul(ug, qmax)',)
        # --- bus ---
        self.v = Var(info='Bus voltage',
                     name='v', tex_name=r'v',
                     unit='p.u.',
                     model='Bus', src='v')
        self.vsq = Var(info='square of Bus voltage',
                       name='vsq', tex_name=r'v^{2}', unit='p.u.',
                       model='Bus',)
        self.vu = Constraint(name='vu',
                             info='Voltage upper limit',
                             e_str='vsq - vmax**2',
                             is_eq=False,)
        self.vl = Constraint(name='vl',
                             info='Voltage lower limit',
                             e_str='-vsq + vmin**2',
                             is_eq=False,)
        # --- line ---
        self.qlf = Var(info='line reactive power',
                       name='qlf', tex_name=r'q_{lf}',
                       unit='p.u.', model='Line',)
        self.lvd = Constraint(info='line voltage drop',
                              name='lvd', is_eq=True,
                              e_str='CftT@vsq - (r * plf + x * qlf)',)
        # --- power balance ---
        # NOTE: following Eqn seems to be wrong, double check
        # g_Q(\Theta, V, Q_g) = B_{bus}V\Theta + Q_{bus,shift} + Q_d + B_{sh} - C_gQ_g = 0
        self.qb = Constraint(info='reactive power balance',
                             name='qb', is_eq=True,
                             e_str='sum(qd) - sum(qg)',)

        # --- objective ---
        # NOTE: no need to revise objective function

    def _post_solve(self):
        self.v.optz.value = np.sqrt(self.vsq.optz.value)
        return super()._post_solve()


class DOPFVIS(DOPF):
    """
    Linearzied distribution OPF with variables for virtual inertia and damping from from REGCV1,
    where power loss are ignored.

    UNDER DEVELOPMENT!

    Reference:

    [1] L. Bai, J. Wang, C. Wang, C. Chen, and F. Li, “Distribution Locational Marginal Pricing (DLMP)
    for Congestion Management and Voltage Support,” IEEE Trans. Power Syst., vol. 33, no. 4,
    pp. 4061–4073, Jul. 2018, doi: 10.1109/TPWRS.2017.2767632.
    """

    def __init__(self, system, config):
        DOPF.__init__(self, system, config)

        # --- params ---
        self.cm = RParam(info='Virtual inertia cost',
                         name='cm', src='cm',
                         tex_name=r'c_{m}', unit=r'$/s',
                         model='VSGCost',
                         indexer='reg', imodel='VSG')
        self.cd = RParam(info='Virtual damping cost',
                         name='cd', src='cd',
                         tex_name=r'c_{d}', unit=r'$/(p.u.)',
                         model='VSGCost',
                         indexer='reg', imodel='VSG',)
        # --- vars ---
        self.M = Var(info='Emulated startup time constant (M=2H) from REGCV1',
                     name='M', tex_name=r'M', unit='s',
                     model='VSG',)
        self.D = Var(info='Emulated damping coefficient from REGCV1',
                     name='D', tex_name=r'D', unit='p.u.',
                     model='VSG',)
        obj = 'sum(c2 * pg**2 + c1 * pg + ug * c0 + cm * M + cd * D)'
        self.obj = Objective(name='tc',
                             info='total cost', unit='$',
                             e_str=obj,
                             sense='min',)



"""
Test file input/output.
"""

import unittest
import numpy as np

import ams


class TestMATPOWER(unittest.TestCase):

    def setUp(self):
        self.mpc5 = ams.io.matpower.m2mpc(ams.get_case('matpower/case5.m'))
        self.mpc14 = ams.io.matpower.m2mpc(ams.get_case('matpower/case14.m'))

    def test_m2mpc(self):
        self.assertTupleEqual(self.mpc5['gencost'].shape, (5, 6))
        self.assertTupleEqual(self.mpc14['gencost'].shape, (5, 7))

    def test_mpc2system(self):
        system5 = ams.system.System()
        ams.io.matpower.mpc2system(self.mpc5, system5)
        # In case5.m, the gencost has type 2 cost model, with 2 parameters.
        np.testing.assert_array_equal(system5.GCost.c2.v,
                                      np.zeros(system5.StaticGen.n))

        system14 = ams.system.System()
        ams.io.matpower.mpc2system(self.mpc14, system14)
        # In case14.m, the gencost has type 2 cost model, with 3 parameters.
        np.testing.assert_array_less(np.zeros(system14.StaticGen.n),
                                     system14.GCost.c2.v,)



import unittest
import numpy as np

import ams


class TestModelMethods(unittest.TestCase):
    """
    Test methods of Model.
    """

    def test_model_set(self):
        """
        Test `Model.set()` method.
        """

        ss = ams.run(
            ams.get_case("ieee14/ieee14.json"),
            default_config=True,
            no_output=True,
        )

        # set a single value
        ss.PQ.set("p0", "PQ_1", "v", 0.25)
        self.assertEqual(ss.PQ.p0.v[0], 0.25)

        # set a list of values
        ss.PQ.set("p0", ["PQ_1", "PQ_2"], "v", [0.26, 0.51])
        np.testing.assert_equal(ss.PQ.p0.v[[0, 1]], [0.26, 0.51])

        # set a list of values
        ss.PQ.set("p0", ["PQ_3", "PQ_5"], "v", [0.52, 0.16])
        np.testing.assert_equal(ss.PQ.p0.v[[2, 4]], [0.52, 0.16])

        # set a list of idxes with a single element to an array of values
        ss.PQ.set("p0", ["PQ_4"], "v", np.array([0.086]))
        np.testing.assert_equal(ss.PQ.p0.v[3], 0.086)

        # set an array of idxes with a single element to an array of values
        ss.PQ.set("p0", np.array(["PQ_4"]), "v", np.array([0.096]))
        np.testing.assert_equal(ss.PQ.p0.v[3], 0.096)

        # set an array of idxes with a list of single value
        ss.PQ.set("p0", np.array(["PQ_4"]), "v", 0.097)
        np.testing.assert_equal(ss.PQ.p0.v[3], 0.097)



# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Used to set and retrieve a PYPOWER options vector.
"""

import numpy as np

from ams.shared import inf


PF_OPTIONS = [
    ('pf_alg', 1, '''power flow algorithm:
1 - Newton's method,
2 - Fast-Decoupled (XB version),
3 - Fast-Decoupled (BX version),
4 - Gauss Seidel'''),

    ('pf_tol', 1e-8, 'termination tolerance on per unit P & Q mismatch'),

    ('pf_max_it', 10, 'maximum number of iterations for Newton\'s method'),

    ('pf_max_it_fd', 30, 'maximum number of iterations for fast '
     'decoupled method'),

    ('pf_max_it_gs', 1000, 'maximum number of iterations for '
     'Gauss-Seidel method'),

    ('enforce_q_lims', False, 'enforce gen reactive power limits, at '
     'expense of |V|'),

    ('pf_dc', False, '''use DC power flow formulation, for power flow and OPF:
False - use AC formulation & corresponding algorithm opts,
True  - use DC formulation, ignore AC algorithm options''')
]

CPF_OPTIONS = [
    ('cpf_parameterization', 3, '''choice of parameterization:
1 - natural,
2 - arc length,
3 - pseudo arc length'''),

    ('cpf_stop_at', 'NOSE', '''determins stopping criterion:
'NOSE' - stop when nose point is reached,
'FULL' - trace full nose curve,
<lam_stop> - stop upon reaching specified target lambda value'''),

    ('cpf_step', 0.05, 'continuation power flow step size'),

    ('cpf_adapt_step', False, '''toggle adaptive step size feature:
False - adaptive step size disabled,
True - adaptive step size enabled'''),

    ('cpf_error_tol', 1e-3, 'tolerance for adaptive step control'),

    ('cpf_step_min', 1e-4, 'minimum allowed step size'),

    ('cpf_step_max', 0.2, 'maximum allowed step size'),

    ('cpf_plot_level', 0, '''control plotting of noze curve:
0 - do not plot nose curve,
1 - plot when completed,
2 - plot incrementally at each iteration,
3 - same as 2, with 'pause' at each iteration'''),

    ('cpf_plot_bus', '', 'index of bus whose voltage is to be plotted'),

    ('cpf_user_callback', '', """string or cell array of strings
with names of user callback functions see 'help cpf_default_callback'"""),

    ('cpf_user_callback_args', '', 'struct passed to user-defined callback functions')
]

OPF_OPTIONS = [
    ('opf_alg', 0, '''algorithm to use for OPF:
0 - choose best default solver available in the
following order, 500, 540, 520 then 100/200
Otherwise the first digit specifies the problem
formulation and the second specifies the solver,
as follows, (see the User's Manual for more details)
500 - generalized formulation, MINOS,
540 - generalized formulation, MIPS
primal/dual interior point method,
545 - generalized formulation (except CCV), SC-MIPS
step-controlled primal/dual interior point method'''),

    #    ('opf_poly2pwl_pts', 10, 'number of evaluation points to use when '
    #     'converting from polynomial to piece-wise linear costs)'),

    ('opf_violation', 5e-6, 'constraint violation tolerance'),

    ('opf_flow_lim', 0, '''qty to limit for branch flow constraints:
0 - apparent power flow (limit in MVA),
1 - active power flow (limit in MW),
2 - current magnitude (limit in MVA at 1 p.u. voltage'''),

    ('opf_ignore_ang_lim', False, 'ignore angle difference limits for '
     'branches even if specified'),

    ('opf_alg_dc', 0, '''solver to use for DC OPF:
0 - choose default solver based on availability in the
following order, 600, 500, 200.
200 - PIPS, Python Interior Point Solver
primal/dual interior point method,
250 - PIPS-sc, step-controlled variant of PIPS
400 - IPOPT, requires pyipopt interface to IPOPT solver
available from: https://projects.coin-or.org/Ipopt/
500 - CPLEX, requires Python interface to CPLEX solver
600 - MOSEK, requires Python interface to MOSEK solver
available from: http://www.mosek.com/
700 - GUROBI, requires Python interface to Gurobi optimizer
available from: http://www.gurobi.com/''')
]

OUTPUT_OPTIONS = [
    ('verbose', 1, '''amount of progress info printed:
0 - print no progress info,
1 - print a little progress info,
2 - print a lot of progress info,
3 - print all progress info'''),

    ('out_all', -1, '''controls printing of results:
-1 - individual flags control what prints,
0 - don't print anything
    (overrides individual flags),
1 - print everything
    (overrides individual flags)'''),

    ('out_sys_sum', True, 'print system summary'),

    ('out_area_sum', False, 'print area summaries'),

    ('out_bus', True, 'print bus detail'),

    ('out_branch', True, 'print branch detail'),

    ('out_gen', False, '''print generator detail
(OUT_BUS also includes gen info)'''),

    ('out_all_lim', -1, '''control constraint info output:
-1 - individual flags control what constraint info prints,
0 - no constraint info (overrides individual flags),
1 - binding constraint info (overrides individual flags),
2 - all constraint info (overrides individual flags)'''),

    ('out_v_lim', 1, '''control output of voltage limit info:
0 - don't print,
1 - print binding constraints only,
2 - print all constraints
(same options for OUT_LINE_LIM, OUT_PG_LIM, OUT_QG_LIM)'''),

    ('out_line_lim', 1, 'control output of line limit info'),

    ('out_pg_lim', 1, 'control output of gen P limit info'),

    ('out_qg_lim', 1, 'control output of gen Q limit info'),

    #    ('out_raw', False, 'print raw data'),

    ('return_raw_der', 0, '''return constraint and derivative info
in results['raw'] (in keys g, dg, df, d2f))''')
]

PDIPM_OPTIONS = [
    ('pdipm_feastol', 0, '''feasibility (equality) tolerance
for Primal-Dual Interior Points Methods, set
to value of OPF_VIOLATION by default'''),
    ('pdipm_gradtol', 1e-6, '''gradient tolerance for
Primal-Dual Interior Points Methods'''),
    ('pdipm_comptol', 1e-6, '''complementary condition (inequality)
tolerance for Primal-Dual Interior Points Methods'''),
    ('pdipm_costtol', 1e-6, '''optimality tolerance for
Primal-Dual Interior Points Methods'''),
    ('pdipm_max_it',  150, '''maximum number of iterations for
Primal-Dual Interior Points Methods'''),
    ('scpdipm_red_it', 20, '''maximum number of reductions per iteration
for Step-Control Primal-Dual Interior Points Methods''')
]

GUROBI_OPTIONS = [
    ('grb_method', 1, '''solution algorithm (Method)
0 - primal simplex
1 - dual simplex
2 - barrier
3 - concurrent (LP only)
4 - deterministic concurrent (LP only)
'''),
    ('grb_timelimit', inf, 'maximum time allowed for solver (TimeLimit)'),
    ('grb_threads', 0, '(auto) maximum number of threads to use (Threads)'),
    ('grb_opt', 0, 'See gurobi_options() for details')
]


def ppoption(ppopt=None, **kwargs):
    """
    Set and retrieve a PYPOWER options vector.

    Parameters
    ----------
    ppopt: dict, optional
        The base options vector. If not provided, the default options vector is used.
    **kwargs: keyword arguments, optional
        New values for the specified options. Each keyword argument represents an option name,
        and its value represents the new value for that option.

    Returns
    -------
    ppopt: dict
        The options vector with the updated values.

    Examples
    --------
    >>> opt = ppoption(PF_ALG=2, PF_TOL=1e-4)
    >>> opt = ppoption(opt, OPF_ALG=565, VERBOSE=2)

    Notes
    -----
    This function is used to set and retrieve PYPOWER options vector.
    The options vector contains various settings and configurations that affect the behavior
    of the PYPOWER routines.

    The `ppopt` parameter, if provided, serves as the base options vector.
    If not provided, the function uses the default options vector.

    Additional options can be specified using keyword arguments, where each keyword
    represents an option name, and its corresponding value represents the new value
    for that option.

    The function returns the options vector with the updated values.
    """

    default_ppopt = {}

    options = PF_OPTIONS + CPF_OPTIONS + OPF_OPTIONS + OUTPUT_OPTIONS + PDIPM_OPTIONS

    for name, default, _ in options:
        default_ppopt[name.upper()] = default

    ppopt = default_ppopt if ppopt == None else ppopt.copy()

    ppopt.update(kwargs)

    return ppopt



"""
Shared constants and delayed imports.

This module is supplementary to the ``andes.shared`` module.
"""
import logging
import unittest
from functools import wraps
from datetime import datetime
from collections import OrderedDict

import cvxpy as cp

from andes.utils.lazyimport import LazyImport

from andes.system import System as adSystem


logger = logging.getLogger(__name__)

sps = LazyImport('import scipy.sparse as sps')
np = LazyImport('import numpy as np')
pd = LazyImport('import pandas as pd')

# --- an empty ANDES system ---
empty_adsys = adSystem()
ad_models = list(empty_adsys.models.keys())

# --- NumPy constants ---
# NOTE: In NumPy 2.0, np.Inf and np.NaN are deprecated.
inf = np.inf
nan = np.nan

# NOTE: copyright
year_end = datetime.now().year
copyright_msg = f'Copyright (C) 2023-{year_end} Jinning Wang'

# NOTE: copied from CVXPY documentation, last checked on 2024/10/30, v1.5
mip_solvers = ['CBC', 'COPT', 'GLPK_MI', 'CPLEX', 'GUROBI',
               'MOSEK', 'SCIP', 'XPRESS', 'SCIPY']

installed_solvers = cp.installed_solvers()

installed_mip_solvers = [s for s in installed_solvers if s in mip_solvers]


def require_MIP_solver(f):
    """
    Decorator for functions that require MIP solver.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if not any(s in mip_solvers for s in installed_solvers):
            raise ModuleNotFoundError("No MIP solver is available.")
        return f(*args, **kwargs)

    return wrapper


def skip_unittest_without_MIP(f):
    """
    Decorator for skipping tests that require MIP solver.
    """
    def wrapper(*args, **kwargs):
        if any(s in mip_solvers for s in installed_solvers):
            pass
        else:
            raise unittest.SkipTest("No MIP solver is available.")
        return f(*args, **kwargs)
    return wrapper


ppc_cols = OrderedDict([
    ('bus', ['bus_i', 'type', 'pd', 'qd', 'gs', 'bs', 'area', 'vm', 'va',
             'baseKV', 'zone', 'vmax', 'vmin', 'lam_p', 'lam_q',
             'mu_vmax', 'mu_vmin']),
    ('branch', ['fbus', 'tbus', 'r', 'x', 'b', 'rate_a', 'rateB_b', 'rate_c',
                'tap', 'shift', 'status', 'angmin',
                'angmax', 'pf', 'qf', 'pt', 'qt', 'mu_sf', 'mu_st',
                'mu_angmin', 'mu_angmax']),
    ('gen', ['bus', 'pg', 'qg', 'qmax', 'qmin', 'vg', 'mbase', 'status',
             'pmax', 'pmin', 'pc1', 'pc2', 'qc1min', 'qc1max', 'qc2min',
             'qc2max', 'ramp_agc', 'ramp_10', 'ramp_30', 'ramp_q',
             'apf', 'mu_pmax', 'mu_pmin', 'mu_qmax', 'mu_qmin']),
    ('gencost', ['model', 'startup', 'shutdown', 'n', 'c2', 'c1', 'c0']),
])


def ppc2df(ppc, model='bus', ppc_cols=ppc_cols):
    """
    Convert PYPOWER dict to pandas DataFrame.

    Parameters
    ----------
    ppc : dict
        PYPOWER case dict.
    model : str
        Model name.
    ppc_cols : dict
        Column names.

    Returns
    -------
    pandas.DataFrame
        DataFrame.

    Examples
    --------
    >>> import ams
    >>> sp = ams.system.example()
    >>> ppc = ams.io.pypower.system2ppc(sp)
    >>> ppc_bus = ams.shared.ppc2df(ppc, 'bus')
    """
    if model not in ppc_cols.keys():
        raise ValueError(f"Invalid model {model}")

    df = pd.DataFrame(ppc[model], columns=ppc_cols[model][0:ppc[model].shape[1]])
    return df



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import os
import ams
import shutil


# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx_panels',
    'IPython.sphinxext.ipython_directive',
    'IPython.sphinxext.ipython_console_highlighting',
    'matplotlib.sphinxext.plot_directive',
    'numpydoc',
    'sphinx_copybutton',
    "nbsphinx",
]

mathjax_path = 'https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-mml-chtml.js'

# Configuration options for plot_directive. See:
# https://github.com/matplotlib/matplotlib/blob/f3ed922d935751e08494e5fb5311d3050a3b637b/lib/matplotlib/sphinxext/plot_directive.py#L81
plot_html_show_source_link = False
plot_html_show_formats = False

# Generate the API documentation when building
autosummary_generate = True
numpydoc_show_class_members = False

templates_path = ['_templates']
# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

exclude_patterns = []

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# General information about the project.
project = 'AMS'
copyright = '2023-2024, Jinning Wang'
author = 'Jinning Wang'

version = ams.__version__
release = ams.__version__

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Use a different latex engine due to possible Unicode characters in the documentation:
# https://docs.readthedocs.io/en/stable/guides/pdf-non-ascii-languages.html
latex_engine = "xelatex"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "use_edit_page_button": True,
    # "footer_items": ["copyright", "sphinx-version", "theme-version"],
}

html_context = {
    "github_url": "https://github.com",
    "github_user": "CURENT",
    "github_repo": "ams",
    "github_version": "master",
    "doc_path": "docs/source",
}

# Tell Jinja2 templates the build is running on Read the Docs
if os.environ.get("READTHEDOCS", "") == "True":
    html_context["READTHEDOCS"] = True

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'ams'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'preamble': r'\DeclareUnicodeCharacter{2588}{-}',
    'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    'pointsize': '11pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'ams.tex', 'AMS Manual',
     'Jinning Wang', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'ams', 'AMS Manual',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'ams', 'AMS Manual',
     author, 'ams', 'Python Software for Scheduling Modeling and Co-Simulation with Dynanics',
     'Miscellaneous'),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),  # Updated URL for numpy
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),  # Updated URL for matplotlib
}

# --- Options for primary sidebar ------------------------------------------
# REF: https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html#layout-sidebar-primary
html_sidebars = {
    # "<page_pattern>": ["list", "of", "templates"]
}

# Favorite icon
html_favicon = 'images/curent.ico'

# Disable smartquotes to display double dashes correctly
smartquotes = False

# import and execute model reference generation script
exec(open("genmodelref.py").read())
exec(open("genroutineref.py").read())

shutil.rmtree("_examples", ignore_errors=True)
shutil.copytree("../../examples", "_examples", )
shutil.rmtree("_examples/demonstration")
shutil.copytree("../../examples/demonstration", "_examples/demo", )

# sphinx-panels shouldn't add bootstrap css since the pydata-sphinx-theme
# already loads it
panels_add_bootstrap_css = False



import unittest

import ams


class TestCodegen(unittest.TestCase):
    """
    Test code generation.
    """

    def test_1_docs(self) -> None:
        sp = ams.system.System()
        out = ''
        for tp in sp.types.values():
            out += tp.doc_all()

        out = ''
        for group in sp.groups.values():
            out += group.doc_all()

    def test_docum(self) -> None:
        sp = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True, no_output=True)
        sp.DCOPF.init()
        docum = sp.DCOPF.docum
        for export in ['plain', 'rest']:
            docum._obj_doc(max_width=78, export=export)
            docum._constr_doc(max_width=78, export=export)
            docum._expr_doc(max_width=78, export=export)
            docum._var_doc(max_width=78, export=export)
            docum._service_doc(max_width=78, export=export)
            docum._param_doc(max_width=78, export=export)
            docum.config.doc(max_width=78, export=export)



import logging

from andes.core.model import ModelData
from andes.utils.tab import Tab
from andes.core.service import BackRef

from ams.core.model import Model

logger = logging.getLogger(__name__)


class RegionData(ModelData):
    def __init__(self):
        super().__init__()


class Region(RegionData, Model):
    """
    Region model for zonal vars.

    Notes
    -----
    1. Region is a collection of buses.
    2. Model ``Region`` is not actually defined in ANDES.

    """

    def __init__(self, system, config):
        RegionData.__init__(self)
        Model.__init__(self, system, config)

        self.group = 'Collection'

        self.Bus = BackRef()
        self.ACTopology = BackRef()

    def bus_table(self):
        """
        Return a formatted table with area idx and bus idx correspondence

        Returns
        -------
        str
            Formatted table

        """
        if self.n:
            header = ['Zone ID', 'Bus ID']
            rows = [(i, j) for i, j in zip(self.idx.v, self.Bus.v)]
            return Tab(header=header, data=rows).draw()
        else:
            return ''



import logging
import unittest

import json

import numpy as np

from andes.shared import rad2deg
import ams
from ams.shared import nan

logger = logging.getLogger(__name__)


class TestKnownResults(unittest.TestCase):

    def setUp(self) -> None:
        with open(ams.get_case('matpower/benchmark.json'), 'r') as file:
            self.mpres = json.load(file)

        self.sp = ams.load(ams.get_case('matpower/case14.m'),
                           setup=True, no_output=True, default_config=True)

    def test_DCPF_case14(self):
        """
        Test DC power flow for case14.
        """
        self.sp.DCPF.run()
        np.testing.assert_allclose(self.sp.DCPF.aBus.v * rad2deg,
                                   np.array(self.mpres['case14']['DCPF']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(self.sp.DCPF.pg.v * self.sp.config.mva,
                                   np.array(self.mpres['case14']['DCPF']['pg']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

    def test_PFlow_case14(self):
        """
        Test power flow for case14.
        """
        self.sp.PFlow.run()
        np.testing.assert_allclose(self.sp.PFlow.aBus.v * rad2deg,
                                   np.array(self.mpres['case14']['PFlow']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.vBus.v,
                                   np.array(self.mpres['case14']['PFlow']['vBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.pg.v.sum() * self.sp.config.mva,
                                   np.array(self.mpres['case14']['PFlow']['pg']).sum(),
                                   rtol=1e-2, atol=1e-2)

    def test_DCOPF_case14(self):
        """
        Test DCOPF for case14.
        """
        self.sp.DCOPF.run(solver='CLARABEL')
        self.assertAlmostEqual(self.sp.DCOPF.obj.v,
                               self.mpres['case14']['DCOPF']['obj'],
                               places=4)
        np.testing.assert_allclose(self.sp.DCOPF.pi.v / self.sp.config.mva,
                                   np.array(self.mpres['case14']['DCOPF']['pi']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

    def test_Matrices_case14(self):
        """
        Test matrices for case14.
        """
        ptdf = self.sp.mats.build_ptdf()
        lodf = self.sp.mats.build_lodf()

        ptdf_mp = load_ptdf(self.mpres, 'case14')
        lodf_mp = load_lodf(self.mpres, 'case14')

        ptdf[np.isnan(ptdf_mp)] = nan
        lodf[np.isnan(lodf_mp)] = nan

        np.testing.assert_allclose(ptdf, ptdf_mp,
                                   equal_nan=True, rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(lodf, lodf_mp,
                                   equal_nan=True, rtol=1e-2, atol=1e-2)


class TestKnownResultsIEEE39(unittest.TestCase):

    def setUp(self) -> None:
        with open(ams.get_case('matpower/benchmark.json'), 'r') as file:
            self.mpres = json.load(file)

        self.sp = ams.load(ams.get_case('matpower/case39.m'),
                           setup=True, no_output=True, default_config=True)

    def test_DCPF_case39(self):
        """
        Test DC power flow for case39.
        """
        self.sp.DCPF.run()
        np.testing.assert_allclose(self.sp.DCPF.aBus.v * rad2deg,
                                   np.array(self.mpres['case39']['DCPF']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(self.sp.DCPF.pg.v.sum() * self.sp.config.mva,
                                   np.array(self.mpres['case39']['DCPF']['pg']).sum(),
                                   rtol=1e-2, atol=1e-2)

    def test_PFlow_case39(self):
        """
        Test power flow for case39.
        """
        self.sp.PFlow.run()
        np.testing.assert_allclose(self.sp.PFlow.aBus.v * rad2deg,
                                   np.array(self.mpres['case39']['PFlow']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.vBus.v,
                                   np.array(self.mpres['case39']['PFlow']['vBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.pg.v.sum() * self.sp.config.mva,
                                   np.array(self.mpres['case39']['PFlow']['pg']).sum(),
                                   rtol=1e-2, atol=1e-2)

    def test_DCOPF_case39(self):
        """
        Test DCOPF for case39.
        """
        self.sp.DCOPF.run(solver='CLARABEL')
        self.assertAlmostEqual(self.sp.DCOPF.obj.v,
                               self.mpres['case39']['DCOPF']['obj'],
                               places=2)
        np.testing.assert_allclose(self.sp.DCOPF.pi.v / self.sp.config.mva,
                                   np.array(self.mpres['case39']['DCOPF']['pi']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

    def test_Matrices_case39(self):
        """
        Test matrices for case39.
        """
        ptdf = self.sp.mats.build_ptdf()
        lodf = self.sp.mats.build_lodf()

        ptdf_mp = load_ptdf(self.mpres, 'case39')
        lodf_mp = load_lodf(self.mpres, 'case39')

        ptdf[np.isnan(ptdf_mp)] = nan
        lodf[np.isnan(lodf_mp)] = nan

        np.testing.assert_allclose(ptdf, ptdf_mp,
                                   equal_nan=True, rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(lodf, lodf_mp,
                                   equal_nan=True, rtol=1e-2, atol=10)


class TestKnownResultsIEEE118(unittest.TestCase):

    def setUp(self) -> None:
        with open(ams.get_case('matpower/benchmark.json'), 'r') as file:
            self.mpres = json.load(file)

        self.sp = ams.load(ams.get_case('matpower/case118.m'),
                           setup=True, no_output=True, default_config=True)

    def test_DCPF_case118(self):
        """
        Test DC power flow for case118.
        """
        self.sp.DCPF.run()
        np.testing.assert_allclose(self.sp.DCPF.aBus.v * rad2deg,
                                   np.array(self.mpres['case118']['DCPF']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(self.sp.DCPF.pg.v.sum() * self.sp.config.mva,
                                   np.array(self.mpres['case118']['DCPF']['pg']).sum(),
                                   rtol=1e-2, atol=1e-2)

    def test_PFlow_case118(self):
        """
        Test power flow for case118.
        """
        self.sp.PFlow.run()
        np.testing.assert_allclose(self.sp.PFlow.aBus.v * rad2deg,
                                   np.array(self.mpres['case118']['PFlow']['aBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.vBus.v,
                                   np.array(self.mpres['case118']['PFlow']['vBus']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(self.sp.PFlow.pg.v.sum() * self.sp.config.mva,
                                   np.array(self.mpres['case118']['PFlow']['pg']).sum(),
                                   rtol=1e-2, atol=1e-2)

    def test_DCOPF_case118(self):
        """
        Test DCOPF for case118.
        """
        self.sp.DCOPF.run(solver='CLARABEL')
        self.assertAlmostEqual(self.sp.DCOPF.obj.v,
                               self.mpres['case118']['DCOPF']['obj'],
                               places=2)
        np.testing.assert_allclose(self.sp.DCOPF.pi.v / self.sp.config.mva,
                                   np.array(self.mpres['case118']['DCOPF']['pi']).reshape(-1),
                                   rtol=1e-2, atol=1e-2)

    def test_Matrices_case118(self):
        """
        Test matrices for case118.
        """
        ptdf = self.sp.mats.build_ptdf()
        lodf = self.sp.mats.build_lodf()

        ptdf_mp = load_ptdf(self.mpres, 'case118')
        lodf_mp = load_lodf(self.mpres, 'case118')

        ptdf[np.isnan(ptdf_mp)] = nan
        lodf[np.isnan(lodf_mp)] = nan

        np.testing.assert_allclose(ptdf, ptdf_mp,
                                   equal_nan=True, rtol=1e-2, atol=1e-2)

        np.testing.assert_allclose(lodf, lodf_mp,
                                   equal_nan=True, rtol=1e-2, atol=10)


def load_ptdf(mpres, case):
    """
    Load PTDF from mpres.

    Parameters
    ----------
    mpres : dict
        The result dictionary.
    case : str
        The case name.

    Returns
    -------
    ptdf : np.ndarray
        The PTDF matrix.
    """
    ptdf_data = np.array(mpres[case]['PTDF'])
    ptdf = np.array([[0 if val == "_NaN_" else val for val in row] for row in ptdf_data],
                    dtype=float)
    return ptdf


def load_lodf(mpres, case):
    """
    Load LODF from mpres.

    Parameters
    ----------
    mpres : dict
        The result dictionary.
    case : str
        The case name.

    Returns
    -------
    lodf : np.ndarray
        The LODF matrix.
    """
    lodf_data = np.array(mpres[case]['LODF'])
    lodf = np.array([[nan if val in ["_NaN_", "-_Inf_", "_Inf_"] else val for val in row] for row in lodf_data],
                    dtype=float)
    # NOTE: force the diagonal to be -1
    np.fill_diagonal(lodf, -1)
    return lodf



"""
MATPOWER parser.
"""
import logging
import numpy as np

from andes.io.matpower import m2mpc
from andes.shared import deg2rad, rad2deg

logger = logging.getLogger(__name__)


def testlines(infile):
    """
    Test if this file is in the MATPOWER format.

    NOT YET IMPLEMENTED.
    """

    return True  # hard coded


def read(system, file):
    """
    Read a MATPOWER data file into mpc, and build AMS device elements.
    """

    mpc = m2mpc(file)
    return mpc2system(mpc, system)


def mpc2system(mpc: dict, system) -> bool:
    """
    Load an mpc dict into an empty AMS system.

    Revised from ``andes.io.matpower.mpc2system``.

    Note that `mbase` in mpc is converted to `Sn`, but it is not actually used in
    MATPOWER nor AMS.

    Parameters
    ----------
    system : ams.system.System
        Empty system to load the data into.
    mpc : dict
        mpc struct names : numpy arrays

    Returns
    -------
    bool
        True if successful, False otherwise.
    """

    # list of buses with slack gen
    sw = []

    system.config.mva = base_mva = mpc['baseMVA']

    for data in mpc['bus']:
        # idx  ty   pd   qd  gs  bs  area  vmag  vang  baseKV  zone  vmax  vmin
        # 0    1    2    3   4   5   6     7     8     9       10    11    12
        idx = int(data[0])
        ty = data[1]
        if ty == 3:
            sw.append(idx)
        pd = data[2] / base_mva
        qd = data[3] / base_mva
        gs = data[4] / base_mva
        bs = data[5] / base_mva
        area = data[6]
        vmag = data[7]
        vang = data[8] * deg2rad
        baseKV = data[9]
        if baseKV == 0:
            baseKV = 110
        zone = data[10]
        vmax = data[11]
        vmin = data[12]

        system.add('Bus', idx=idx, name=None,
                   type=ty, Vn=baseKV,
                   v0=vmag, a0=vang,
                   vmax=vmax, vmin=vmin,
                   area=area, zone=zone)
        if pd != 0 or qd != 0:
            system.add('PQ', bus=idx, name=None, Vn=baseKV, p0=pd, q0=qd)
        if gs or bs:
            system.add('Shunt', bus=idx, name=None, Vn=baseKV, g=gs, b=bs)

    gen_idx = 0
    if mpc['gen'].shape[1] <= 10:  # missing data
        mpc_gen = np.zeros((mpc['gen'].shape[0], 21), dtype=np.float64)
        mpc_gen[:, :10] = mpc['gen']
        mpc_gen[:, 16] = system.PV.Ragc.default * base_mva / 60
        mpc_gen[:, 17] = system.PV.R10.default * base_mva
        mpc_gen[:, 18] = system.PV.R30.default * base_mva
        mpc_gen[:, 19] = system.PV.Rq.default * base_mva / 60
    else:
        mpc_gen = mpc['gen']
    for data in mpc_gen:
        # bus  pg  qg  qmax  qmin  vg  mbase  status  pmax  pmin
        # 0    1   2   3     4     5   6      7       8     9
        # pc1  pc2  qc1min  qc1max  qc2min  qc2max  ramp_agc  ramp_10
        # 10   11   12      13      14      15      16        17
        # ramp_30  ramp_q  apf
        # 18       19      20
        bus_idx = int(data[0])
        gen_idx += 1
        vg = data[5]
        status = int(data[7])
        pg = data[1] / base_mva
        qg = data[2] / base_mva
        qmax = data[3] / base_mva
        qmin = data[4] / base_mva
        pmax = data[8] / base_mva
        pmin = data[9] / base_mva
        pc1 = data[10] / base_mva
        pc2 = data[11] / base_mva
        qc1min = data[12] / base_mva
        qc1max = data[13] / base_mva
        qc2min = data[14] / base_mva
        qc2max = data[15] / base_mva
        ramp_agc = 60 * data[16] / base_mva  # MW/min -> MW/h
        ramp_10 = data[17] / base_mva  # MW -> MW/h
        ramp_30 = data[18] / base_mva  # MW -> MW/h
        ramp_q = 60 * data[19] / base_mva  # MVAr/min -> MVAr/h
        apf = data[20]

        uid = system.Bus.idx2uid(bus_idx)
        vn = system.Bus.Vn.v[uid]
        a0 = system.Bus.a0.v[uid]

        if bus_idx in sw:
            system.add('Slack', idx=gen_idx, bus=bus_idx, busr=bus_idx,
                       name=None,
                       u=status, Sn=data[6],
                       Vn=vn, v0=vg, p0=pg, q0=qg, a0=a0,
                       pmax=pmax, pmin=pmin,
                       qmax=qmax, qmin=qmin,
                       Pc1=pc1, Pc2=pc2,
                       Qc1min=qc1min, Qc1max=qc1max,
                       Qc2min=qc2min, Qc2max=qc2max,
                       Ragc=ramp_agc, R10=ramp_10,
                       R30=ramp_30, Rq=ramp_q,
                       apf=apf)
        else:
            system.add('PV', idx=gen_idx, bus=bus_idx, busr=bus_idx,
                       name=None,
                       u=status, Sn=data[6],
                       Vn=vn, v0=vg, p0=pg, q0=qg,
                       pmax=pmax, pmin=pmin,
                       qmax=qmax, qmin=qmin,
                       Pc1=pc1, Pc2=pc2,
                       Qc1min=qc1min, Qc1max=qc1max,
                       Qc2min=qc2min, Qc2max=qc2max,
                       Ragc=ramp_agc, R10=ramp_10,
                       R30=ramp_30, Rq=ramp_q,
                       apf=apf)

    for data in mpc['branch']:
        # fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle
        # 0     1       2   3   4   5       6       7       8       9
        # status	angmin	angmax	Pf	Qf	Pt	Qt
        # 10        11      12      13  14  15  16
        fbus = int(data[0])
        tbus = int(data[1])
        r = data[2]
        x = data[3]
        b = data[4]
        rate_a = data[5] / base_mva
        rate_b = data[6] / base_mva
        rate_c = data[7] / base_mva
        amin = data[11] * deg2rad
        amax = data[12] * deg2rad

        status = int(data[10])

        if (data[8] == 0.0) or (data[8] == 1.0 and data[9] == 0.0):
            # not a transformer
            tf = False
            ratio = 1
            angle = 0
        else:
            tf = True
            ratio = data[8]
            angle = data[9] * deg2rad

        vf = system.Bus.Vn.v[system.Bus.idx2uid(fbus)]
        vt = system.Bus.Vn.v[system.Bus.idx2uid(tbus)]
        system.add('Line', u=status, name=f'Line {fbus:.0f}-{tbus:.0f}',
                   Vn1=vf, Vn2=vt,
                   bus1=fbus, bus2=tbus,
                   r=r, x=x, b=b,
                   trans=tf, tap=ratio, phi=angle,
                   rate_a=rate_a, rate_b=rate_b, rate_c=rate_c,
                   amin=amin, amax=amax)

    if ('bus_name' in mpc) and (len(mpc['bus_name']) == len(system.Bus.name.v)):
        system.Bus.name.v[:] = mpc['bus_name']

    # --- gencost ---
    if 'gencost' in mpc:
        gcost_idx = 0
        gen_idx = np.arange(mpc['gen'].shape[0]) + 1
        mpc_cost = mpc['gencost']
        for data, gen in zip(mpc_cost, gen_idx):
            # NOTE: only type 2 costs are supported for now
            # type  startup shutdown	n	c2  c1  c0
            # 0     1       2           3   4   5   6
            if data[0] != 2:
                raise ValueError('Only MODEL 2 costs are supported')
            gcost_idx += 1
            gctype = int(data[0])
            startup = data[1]
            shutdown = data[2]
            if data[3] == 3:
                c2 = data[4] * base_mva ** 2
                c1 = data[5] * base_mva
                c0 = data[6]
            elif data[3] == 2:
                c2 = 0
                c1 = data[4] * base_mva
                c0 = data[5]
            else:
                raise ValueError('Unrecognized gencost model, please use eighter quadratic or linear cost model')
            system.add('GCost', gen=int(gen),
                       u=1, type=gctype,
                       idx=gcost_idx,
                       name=None,
                       csu=startup, csd=shutdown,
                       c2=c2, c1=c1, c0=c0
                       )

    # --- region ---
    zone_id = np.unique(system.Bus.zone.v).astype(int)
    for zone in zone_id:
        zone_idx = f'ZONE_{zone}'
        system.add('Region', idx=zone_idx, name=None)
    bus_zone = system.Bus.zone.v
    bus_zone = [f'ZONE_{int(zone)}' for zone in bus_zone]
    system.Bus.zone.v = bus_zone
    return True


def _get_bus_id_caller(bus):
    """
    Helper function to get the bus id. Force bus id to be uid+1.

    Parameters
    ----------
    bus : andes.models.bus.Bus
        Bus object

    Returns
    -------
    lambda function to that takes bus idx and returns bus id for matpower case
    """

    if np.array(bus.idx.v).dtype in ['int', 'float']:
        return lambda x: x
    else:
        return lambda x: list(np.array(bus.idx2uid(x)) + 1)


def system2mpc(system) -> dict:
    """
    Convert data from an AMS system to an mpc dict.

    In the ``gen`` section, slack generators preceeds PV generators.

    Compared to the ``andes.io.matpower.system2mpc``,
    this function includes the generator cost data in the ``gencost``
    section.
    Additionally, ``c2`` and ``c1`` are scaled by ``base_mva`` to match
    MATPOWER unit ``MW``.

    Parameters
    ----------
    system : ams.core.system.System
        AMS system

    Returns
    -------
    mpc: dict
        MATPOWER mpc dict
    """

    mpc = dict(version='2',
               baseMVA=system.config.mva,
               bus=np.zeros((system.Bus.n, 13), dtype=np.float64),
               gen=np.zeros((system.PV.n + system.Slack.n, 21), dtype=np.float64),
               branch=np.zeros((system.Line.n, 17), dtype=np.float64),
               gencost=np.zeros((system.GCost.n, 7), dtype=np.float64),
               )

    if system.Bus.name.v is not None:
        mpc['bus_name'] = system.Bus.name.v

    base_mva = system.config.mva

    # --- bus ---
    bus = mpc['bus']
    gen = mpc['gen']

    to_busid = _get_bus_id_caller(system.Bus)

    bus[:, 0] = to_busid(system.Bus.idx.v)
    bus[:, 1] = 1
    bus[:, 7] = system.Bus.v0.v
    bus[:, 8] = system.Bus.a0.v * rad2deg
    bus[:, 9] = system.Bus.Vn.v
    bus[:, 11] = system.Bus.vmax.v
    bus[:, 12] = system.Bus.vmin.v

    # --- zone ---
    ZONE_I = system.Region.idx.v
    if len(ZONE_I) > 0:
        mapping = {busi0: i for i, busi0 in enumerate(ZONE_I)}
        bus[:, 10] = np.array([mapping[busi0] for busi0 in system.Bus.zone.v])

    # --- PQ ---
    if system.PQ.n > 0:
        pq_pos = system.Bus.idx2uid(system.PQ.bus.v)
        bus[pq_pos, 2] = system.PQ.p0.v * base_mva
        bus[pq_pos, 3] = system.PQ.q0.v * base_mva

    # --- Shunt ---
    if system.Shunt.n > 0:
        shunt_pos = system.Bus.idx2uid(system.Shunt.bus.v)
        bus[shunt_pos, 4] = system.Shunt.g.v * base_mva
        bus[shunt_pos, 5] = system.Shunt.b.v * base_mva

    # --- PV ---
    if system.PV.n > 0:
        PV = system.PV
        pv_pos = system.Bus.idx2uid(PV.bus.v)
        bus[pv_pos, 1] = 2
        gen[system.Slack.n:, 0] = to_busid(PV.bus.v)
        gen[system.Slack.n:, 1] = PV.p0.v * base_mva
        gen[system.Slack.n:, 2] = PV.q0.v * base_mva
        gen[system.Slack.n:, 3] = PV.qmax.v * base_mva
        gen[system.Slack.n:, 4] = PV.qmin.v * base_mva
        gen[system.Slack.n:, 5] = PV.v0.v
        gen[system.Slack.n:, 6] = PV.Sn.v
        gen[system.Slack.n:, 7] = PV.u.v
        gen[system.Slack.n:, 8] = (PV.ctrl.v * PV.pmax.v + (1 - PV.ctrl.v) * PV.p0.v) * base_mva
        gen[system.Slack.n:, 9] = (PV.ctrl.v * PV.pmin.v + (1 - PV.ctrl.v) * PV.p0.v) * base_mva
        gen[system.Slack.n:, 16] = PV.Ragc.v * base_mva * 60    # MW/h -> MW/min
        gen[system.Slack.n:, 17] = PV.R10.v * base_mva
        gen[system.Slack.n:, 18] = PV.R30.v * base_mva
        gen[system.Slack.n:, 19] = PV.Rq.v * base_mva * 60  # MVAr/h -> MVAr/min

    # --- Slack ---
    if system.Slack.n > 0:
        slack_pos = system.Bus.idx2uid(system.Slack.bus.v)
        bus[slack_pos, 1] = 3
        bus[slack_pos, 8] = system.Slack.a0.v * rad2deg
        gen[:system.Slack.n, 0] = to_busid(system.Slack.bus.v)
        gen[:system.Slack.n, 1] = system.Slack.p0.v * base_mva
        gen[:system.Slack.n, 2] = system.Slack.q0.v * base_mva
        gen[:system.Slack.n, 3] = system.Slack.qmax.v * base_mva
        gen[:system.Slack.n, 4] = system.Slack.qmin.v * base_mva
        gen[:system.Slack.n, 5] = system.Slack.v0.v
        gen[:system.Slack.n, 6] = system.Slack.Sn.v
        gen[:system.Slack.n, 7] = system.Slack.u.v
        gen[:system.Slack.n, 8] = system.Slack.pmax.v * base_mva
        gen[:system.Slack.n, 9] = system.Slack.pmin.v * base_mva
        gen[:system.Slack.n, 16] = system.Slack.Ragc.v * base_mva * 60    # MW/h -> MW/min
        gen[:system.Slack.n, 17] = system.Slack.R10.v * base_mva
        gen[:system.Slack.n, 18] = system.Slack.R30.v * base_mva
        gen[:system.Slack.n, 19] = system.Slack.Rq.v * base_mva * 60    # MVAr/h -> MVAr/min

    if system.Line.n > 0:
        branch = mpc['branch']
        branch[:, 0] = to_busid(system.Line.bus1.v)
        branch[:, 1] = to_busid(system.Line.bus2.v)
        branch[:, 2] = system.Line.r.v
        branch[:, 3] = system.Line.x.v
        branch[:, 4] = system.Line.b.v
        branch[:, 5] = system.Line.rate_a.v * base_mva
        branch[:, 6] = system.Line.rate_b.v * base_mva
        branch[:, 7] = system.Line.rate_c.v * base_mva
        branch[:, 8] = system.Line.tap.v
        branch[:, 9] = system.Line.phi.v * rad2deg
        branch[:, 10] = system.Line.u.v
        branch[:, 11] = system.Line.amin.v * rad2deg
        branch[:, 12] = system.Line.amax.v * rad2deg

    # --- GCost ---
    # NOTE: adjust GCost sequence to match the generator sequence
    if system.GCost.n > 0:
        stg_idx = system.Slack.idx.v + system.PV.idx.v
        gcost_idx = system.GCost.find_idx(keys=['gen'], values=[stg_idx])
        gcost_uid = system.GCost.idx2uid(gcost_idx)
        gencost = mpc['gencost']
        gencost[:, 0] = system.GCost.type.v[gcost_uid]
        gencost[:, 1] = system.GCost.csu.v[gcost_uid]
        gencost[:, 2] = system.GCost.csd.v[gcost_uid]
        gencost[:, 3] = 3
        gencost[:, 4] = system.GCost.c2.v[gcost_uid] / base_mva / base_mva
        gencost[:, 5] = system.GCost.c1.v[gcost_uid] / base_mva
        gencost[:, 6] = system.GCost.c0.v[gcost_uid]
    else:
        mpc.pop('gencost')

    return mpc



import logging  # NOQA

from andes.models.group import GroupBase as andes_GroupBase
from andes.core.service import BackRef

logger = logging.getLogger(__name__)


class GroupBase(andes_GroupBase):
    """
    Base class for groups.

    Add common_vars and common_params to the group class.

    This class is revised from ``andes.models.group.GroupBase``.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('idx',))

    def get_idx(self):
        """
        Return the value of group idx sorted in a human-readable style.

        Notes
        -----
        This function sorts the idx values using a custom sorting key,
        which handles varying length strings with letters and numbers.
        """
        all_idx = [mdl.idx.v for mdl in self.models.values()]

        # Custom sorting function to handle varying length strings with letters and numbers
        def custom_sort_key(item):
            try:
                return int(item)  # Try to convert to integer for pure numeric strings
            except ValueError:
                try:
                    # Extract numeric part for strings with letters and numbers
                    return int(''.join(filter(str.isdigit, item)))
                except ValueError:
                    return item  # Return as is if not numeric

        group_idx = [sorted(mdl_idx, key=custom_sort_key) for mdl_idx in all_idx]
        flat_list = [item for sublist in group_idx for item in sublist]
        return flat_list

    def _check_src(self, src: str):
        """
        Helper function for checking if ``src`` is a shared field.

        The requirement is not strictly enforced and is only for debugging purposed.

        Disable debug logging in dispath modeling.
        """
        if src not in self.common_vars + self.common_params:
            logger.debug(f'Group <{self.class_name}> does not share property <{src}>.')

    def __repr__(self):
        dev_text = 'device' if self.n == 1 else 'devices'
        return f'{self.class_name} ({self.n} {dev_text}) at {hex(id(self))}'

    def __setattr__(self, key, value):
        if hasattr(value, 'owner'):
            if value.owner is None:
                value.owner = self
        if hasattr(value, 'name'):
            if value.name is None:
                value.name = key

        if isinstance(value, BackRef):
            self.services_ref[key] = value

        super().__setattr__(key, value)

    def set_backref(self, name, from_idx, to_idx):
        """
        Set idxes to ``BackRef``, and set them to models.
        """

        uid = self.idx2uid(to_idx)
        self.services_ref[name].v[uid].append(from_idx)

        model = self.idx2model(to_idx)
        model.set_backref(name, from_idx, to_idx)


class Undefined(GroupBase):
    """
    The undefined group. Holds models with no ``group``.
    """

    def __init__(self):
        super().__init__()


class ACTopology(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_vars.extend(('a', 'v'))


class RenGen(GroupBase):
    """
    Renewable generator (converter) group.

    See ANDES Documentation SynGen here for the notes on replacing StaticGen and setting the power
    ratio parameters.

    Reference:

    [1] ANDES Documentation, RenGen, [Online]

    Available:

    https://docs.andes.app/en/latest/groupdoc/RenGen.html#rengen
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'gen', 'Sn', 'q0'))
        self.common_vars.extend(('Pe', 'Qe'))


class VSG(GroupBase):
    """
    Renewable generator with virtual synchronous generator (VSG) control group.

    Note that this is a group separate from ``RenGen`` for VSG scheduling study.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'gen', 'Sn'))
        self.common_vars.extend(('Pe', 'Qe'))


class DG(GroupBase):
    """
    Distributed generation (small-scale).

    See ANDES Documentation SynGen here for the notes on replacing StaticGen and setting the power
    ratio parameters.

    Reference:

    [1] ANDES Documentation, SynGen, [Online]

    Available:

    https://docs.andes.app/en/latest/groupdoc/SynGen.html#syngen
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'fn'))


class Cost(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_params.extend(('gen',))


class Collection(GroupBase):
    """
    Collection of topology models
    """

    def __init__(self):
        super().__init__()


class Horizon(GroupBase):
    """
    Time horizon group.
    """

    def __init__(self):
        super().__init__()


class Reserve(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_params.extend(('zone',))


class StaticGen(GroupBase):
    """
    Generator group.

    Notes
    -----
    For co-simulation with ANDES, check
    `ANDES StaticGen <https://docs.andes.app/en/latest/groupdoc/StaticGen.html#staticgen>`_
    for replacing static generators with dynamic generators.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'Sn', 'Vn', 'p0', 'q0', 'ra', 'xs', 'subidx',
                                   'pmax', 'pmin', 'pg0', 'ctrl', 'R10', 'td1', 'td2',
                                   'zone'))
        self.common_vars.extend(('p', 'q'))


class ACLine(GroupBase):
    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus1', 'bus2', 'r', 'x'))


class ACShort(GroupBase):
    def __init__(self):
        super(ACShort, self).__init__()
        self.common_params.extend(('bus1', 'bus2'))


class StaticLoad(GroupBase):
    """
    Static load group.
    """

    def __init__(self):
        super().__init__()
        self.common_params.extend(('bus', 'p0', 'q0', 'ctrl', 'zone'))


class StaticShunt(GroupBase):
    """
    Static shunt compensator group.
    """
    pass


class Information(GroupBase):
    """
    Group for information container models.
    """

    def __init__(self):
        GroupBase.__init__(self)
        self.common_params = []



"""
Module for routine data.
"""

import logging
import os
from typing import Optional, Union, Type, Iterable
from collections import OrderedDict

import numpy as np

from andes.core import Config
from andes.utils.misc import elapsed

from ams.core.param import RParam
from ams.core.symprocessor import SymProcessor
from ams.core.documenter import RDocumenter
from ams.core.service import RBaseService, ValueService
from ams.opt.omodel import OModel, Param, Var, Constraint, Objective, ExpressionCalc

from ams.shared import pd

logger = logging.getLogger(__name__)


class RoutineBase:
    """
    Class to hold descriptive routine models and data mapping.

    Attributes
    ----------
    system : Optional[Type]
        The system object associated with the routine.
    config : Config
        Configuration object for the routine.
    info : Optional[str]
        Information about the routine.
    tex_names : OrderedDict
        LaTeX names for the routine parameters.
    syms : SymProcessor
        Symbolic processor for the routine.
    _syms : bool
        Flag indicating whether symbols have been generated.
    rparams : OrderedDict
        Registry for RParam objects.
    services : OrderedDict
        Registry for service objects.
    params : OrderedDict
        Registry for Param objects.
    vars : OrderedDict
        Registry for Var objects.
    constrs : OrderedDict
        Registry for Constraint objects.
    exprs : OrderedDict
        Registry for Expression objects.
    obj : Optional[Objective]
        Objective of the routine.
    initialized : bool
        Flag indicating whether the routine has been initialized.
    type : str
        Type of the routine.
    docum : RDocumenter
        Documentation generator for the routine.
    map1 : OrderedDict
        Mapping from ANDES.
    map2 : OrderedDict
        Mapping to ANDES.
    om : OModel
        Optimization model for the routine.
    exec_time : float
        Execution time of the routine.
    exit_code : int
        Exit code of the routine.
    converged : bool
        Flag indicating whether the routine has converged.
    converted : bool
        Flag indicating whether AC conversion has been performed.
    """

    def __init__(self, system=None, config=None):
        """
        Initialize the routine.

        Parameters
        ----------
        system : Optional[Type]
            The system object associated with the routine.
        config : Optional[dict]
            Configuration dictionary for the routine.
        """
        self.system = system
        self.config = Config(self.class_name)
        self.info = None
        self.tex_names = OrderedDict(
            (
                ("sys_f", "f_{sys}"),
                ("sys_mva", "S_{b,sys}"),
            )
        )
        self.syms = SymProcessor(self)      # symbolic processor
        self._syms = False                  # symbol generation flag

        self.rparams = OrderedDict()        # RParam registry
        self.services = OrderedDict()       # Service registry
        self.params = OrderedDict()         # Param registry
        self.vars = OrderedDict()           # Var registry
        self.constrs = OrderedDict()        # Constraint registry
        self.exprs = OrderedDict()          # Expression registry
        self.obj = None                     # Objective
        self.initialized = False            # initialization flag
        self.type = "UndefinedType"         # routine type
        self.docum = RDocumenter(self)      # documentation generator

        # --- sync mapping ---
        self.map1 = OrderedDict()  # from ANDES
        self.map2 = OrderedDict()  # to ANDES

        # --- optimization modeling ---
        self.om = OModel(routine=self)      # optimization model

        if config is not None:
            self.config.load(config)

        # NOTE: the difference between exit_code and converged is that
        # exit_code is the solver exit code, while converged is the
        # convergence flag of the routine.
        self.exec_time = 0.0        # running time
        self.exit_code = 0          # exit code
        self.converged = False      # convergence flag
        self.converted = False          # AC conversion flag

    @property
    def class_name(self):
        return self.__class__.__name__

    def get(self, src: str, idx, attr: str = 'v',
            horizon: Optional[Union[int, str, Iterable]] = None):
        """
        Get the value of a variable or parameter.

        Parameters
        ----------
        src: str
            Name of the variable or parameter.
        idx: int, str, or list
            Index of the variable or parameter.
        attr: str
            Attribute name.
        horizon: list, optional
            Horizon index.
        """
        if src not in self.__dict__.keys():
            raise ValueError(f"<{src}> does not exist in <<{self.class_name}>.")
        item = self.__dict__[src]

        if not hasattr(item, attr):
            raise ValueError(f"{attr} does not exist in {self.class_name}.{src}.")

        idx_all = item.get_idx()

        if idx_all is None:
            raise ValueError(f"<{self.class_name}> item <{src}> has no idx.")

        is_format = False  # whether the idx is formatted as a list
        idx_u = None
        if isinstance(idx, (str, int)):
            idx_u = [idx]
            is_format = True
        elif isinstance(idx, (np.ndarray, pd.Series)):
            idx_u = idx.tolist()
        elif isinstance(idx, list):
            idx_u = idx.copy()

        loc = [idx_all.index(idxe) if idxe in idx_all else None for idxe in idx_u]
        if None in loc:
            idx_none = [idxe for idxe in idx_u if idxe not in idx_all]
            msg = f"Var <{self.class_name}.{src}> does not contain value with idx={idx_none}"
            raise ValueError(msg)
        out = getattr(item, attr)[loc]

        if horizon is not None:
            if item.horizon is None:
                raise ValueError(f"horizon is not defined for {self.class_name}.{src}.")
            horizon_all = item.horizon.get_idx()
            if not isinstance(horizon, list):
                raise TypeError(f"horizon must be a list, not {type(horizon)}.")
            loc_h = [
                horizon_all.index(idxe) if idxe in horizon_all else None
                for idxe in horizon
            ]
            if None in loc_h:
                idx_none = [idxe for idxe in horizon if idxe not in horizon_all]
                msg = f"Var <{self.class_name}.{src}> does not contain horizon with idx={idx_none}"
                raise ValueError(msg)
            out = out[:, loc_h]
            if out.shape[1] == 1:
                out = out[:, 0]

        return out[0] if is_format else out

    def set(self, src: str, idx, attr: str = "v", value=0.0):
        """
        Set the value of an attribute of a routine parameter.

        Performs ``self.<src>.<attr>[idx] = value``. This method will not modify
        the input values from the case file that have not been converted to the
        system base. As a result, changes applied by this method will not affect
        the dumped case file.

        To alter parameters and reflect it in the case file, use :meth:`alter`
        instead.

        Parameters
        ----------
        src : str
            Name of the model property
        idx : str, int, float, array-like
            Indices of the devices
        attr : str, optional, default='v'
            The internal attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        value : array-like
            New values to be set

        Returns
        -------
        bool
            True when successful.
        """
        if self.__dict__[src].owner is not None:
            # TODO: fit to `_v` type param in the future
            owner = self.__dict__[src].owner
            src0 = self.__dict__[src].src
            try:
                res = owner.set(src=src0, idx=idx, attr=attr, value=value)
                return res
            except KeyError as e:
                msg = f"Failed to set <{src0}> in <{owner.class_name}>. "
                msg += f"Original error: {e}"
                raise KeyError(msg)
        else:
            # FIXME: add idx for non-grouped variables
            raise TypeError(f"Variable {self.name} has no owner.")

    def doc(self, max_width=78, export="plain"):
        """
        Retrieve routine documentation as a string.
        """
        return self.docum.get(max_width=max_width, export=export)

    def _get_off_constrs(self):
        """
        Chcek if constraints are turned off.
        """
        disabled = []
        for cname, c in self.constrs.items():
            if c.is_disabled:
                disabled.append(cname)
        if len(disabled) > 0:
            msg = "Disabled constraints: "
            d_str = [f'{constr}' for constr in disabled]
            msg += ", ".join(d_str)
            logger.warning(msg)
        return disabled

    def _data_check(self, info=True):
        """
        Check if data is valid for a routine.

        Parameters
        ----------
        info: bool
            Whether to print warning messages.
        """
        logger.debug(f"Entering data check for <{self.class_name}>")
        no_input = []
        owner_list = []
        for rname, rparam in self.rparams.items():
            if rparam.owner is not None:
                # NOTE: skip checking Shunt.g
                if (rparam.owner.class_name == 'Shunt') and (rparam.src == 'g'):
                    pass
                elif rparam.owner.n == 0:
                    no_input.append(rname)
                    owner_list.append(rparam.owner.class_name)
            # TODO: add more data config check?
            if rparam.config.pos:
                if not np.all(rparam.v > 0):
                    logger.warning(f"RParam <{rname}> should have all positive values.")
        if len(no_input) > 0:
            if info:
                msg = f"Following models are missing in input: {set(owner_list)}"
                logger.error(msg)
            return False
        # TODO: add data validation for RParam, typical range, etc.
        logger.debug(" -> Data check passed")
        return True

    def init(self, **kwargs):
        """
        Initialize the routine.

        Other parameters
        ----------------
        force: bool
            Whether to force initialization regardless of the current initialization status.
        force_mats: bool
            Whether to force build the system matrices, goes to `self.system.mats.build()`.
        force_constr: bool
            Whether to turn on all constraints.
        force_om: bool
            Whether to force initialize the optimization model.
        """
        force = kwargs.pop('force', False)
        force_mats = kwargs.pop('force_mats', False)
        force_constr = kwargs.pop('force_constr', False)
        force_om = kwargs.pop('force_om', False)

        skip_all = not (force and force_mats) and self.initialized

        if skip_all:
            logger.debug(f"{self.class_name} has already been initialized.")
            return True

        t0, _ = elapsed()
        # --- data check ---
        self._data_check()

        # --- turn on all constrs ---
        if force_constr:
            for constr in self.constrs.values():
                constr.is_disabled = False

        # --- matrix build ---
        self.system.mats.build(force=force_mats)

        # --- constraint check ---
        _ = self._get_off_constrs()

        if not self.om.initialized:
            self.om.init(force=force_om)
        _, s_init = elapsed(t0)

        msg = f"<{self.class_name}> "
        if self.om.initialized:
            msg += f"initialized in {s_init}."
            self.initialized = True
        else:
            msg += "initialization failed!"
            self.initialized = False
        logger.info(msg)
        return self.initialized

    def solve(self, **kwargs):
        """
        Solve the routine optimization model.
        """
        raise NotImplementedError

    def unpack(self, **kwargs):
        """
        Unpack the results.
        """
        raise NotImplementedError

    def _post_solve(self):
        """
        Post-solve calculation.
        """
        raise NotImplementedError

    def run(self, **kwargs):
        """
        Run the routine.
        args and kwargs go to `self.solve()`.

        Force initialization (`force_init=True`) will do the following:
        - Rebuild the system matrices
        - Enable all constraints
        - Reinitialize the optimization model

        Parameters
        ----------
        force_init: bool
            Whether to force re-initialize the routine.
        force_mats: bool
            Whether to force build the system matrices.
        force_constr: bool
            Whether to turn on all constraints.
        force_om: bool
            Whether to force initialize the OModel.
        """
        # --- setup check ---
        force_init = kwargs.pop('force_init', False)
        force_mats = kwargs.pop('force_mats', False)
        force_constr = kwargs.pop('force_constr', False)
        force_om = kwargs.pop('force_om', False)
        self.init(force=force_init, force_mats=force_mats,
                  force_constr=force_constr, force_om=force_om)

        # --- solve optimization ---
        t0, _ = elapsed()
        _ = self.solve(**kwargs)
        status = self.om.prob.status
        self.exit_code = self.syms.status[status]
        self.converged = self.exit_code == 0
        _, s = elapsed(t0)
        self.exec_time = float(s.split(" ")[0])
        sstats = self.om.prob.solver_stats  # solver stats
        if sstats.num_iters is None:
            n_iter = -1
        else:
            n_iter = int(sstats.num_iters)
        n_iter_str = f"{n_iter} iterations " if n_iter > 1 else f"{n_iter} iteration "
        if self.exit_code == 0:
            msg = f"<{self.class_name}> solved as {status} in {s}, converged in "
            msg += n_iter_str + f"with {sstats.solver_name}."
            logger.warning(msg)
            self.unpack(**kwargs)
            self._post_solve()
            return True
        else:
            msg = f"{self.class_name} failed as {status} in "
            msg += n_iter_str + f"with {sstats.solver_name}!"
            logger.warning(msg)
            return False

    def export_csv(self, path=None):
        """
        Export scheduling results to a csv file.
        For multi-period routines, the column "Time" is the time index of
        ``timeslot.v``, which usually comes from ``EDTSlot`` or ``UCTSlot``.
        The rest columns are the variables registered in ``vars``.

        For single-period routines, the column "Time" have a pseduo value of "T1".

        Parameters
        ----------
        path : str
            path of the csv file to save

        Returns
        -------
        str
            The path of the exported csv file
        """
        if not self.converged:
            logger.warning("Routine did not converge, aborting export.")
            return None
        if not path:
            if self.system.files.fullname is None:
                logger.info("Input file name not detacted. Using `Untitled`.")
                file_name = f'Untitled_{self.class_name}'
            else:
                file_name = os.path.splitext(self.system.files.fullname)[0]
                file_name += f'_{self.class_name}'
            path = os.path.join(os.getcwd(), file_name + '.csv')

        idxes = [var.get_idx() for var in self.vars.values()]
        var_names = [var for var in self.vars.keys()]

        if hasattr(self, 'timeslot'):
            timeslot = self.timeslot.v.copy()
            data_dict = OrderedDict([('Time', timeslot)])
        else:
            timeslot = None
            data_dict = OrderedDict([('Time', 'T1')])

        for var, idx in zip(var_names, idxes):
            header = [f'{var} {dev}' for dev in idx]
            data = self.get(src=var, idx=idx, horizon=timeslot).round(6)
            data_dict.update(OrderedDict(zip(header, data)))

        if timeslot is None:
            data_dict = OrderedDict([(k, [v]) for k, v in data_dict.items()])

        pd.DataFrame(data_dict).to_csv(path, index=False)
        return file_name + '.csv'

    def summary(self, **kwargs):
        """
        Summary interface
        """
        raise NotImplementedError

    def __repr__(self):
        return f"{self.class_name} at {hex(id(self))}"

    def _ppc2ams(self):
        """
        Convert PYPOWER results to AMS.
        """
        raise NotImplementedError

    def dc2ac(self, **kwargs):
        """
        Convert the DC-based results with ACOPF.
        """
        raise NotImplementedError

    def _check_attribute(self, key, value):
        """
        Check the attribute pair for valid names while instantiating the class.

        This function assigns `owner` to the model itself, assigns the name and tex_name.
        """
        if key in self.__dict__:
            existing_keys = []
            for rtn_type in ["constrs", "vars", "rparams", "services"]:
                if rtn_type in self.__dict__:
                    existing_keys += list(self.__dict__[rtn_type].keys())
            if key in existing_keys:
                msg = f"Attribute <{key}> already exists in <{self.class_name}>."
                logger.warning(msg)

        # register owner routine instance of following attributes
        if isinstance(value, (RBaseService)):
            value.rtn = self

    def __setattr__(self, key, value):
        """
        Overload the setattr function to register attributes.

        Parameters
        ----------
        key: str
            name of the attribute
        value:
            value of the attribute
        """

        # NOTE: value.id is not in use yet
        if isinstance(value, Var):
            value.id = len(self.vars)
        self._check_attribute(key, value)
        self._register_attribute(key, value)

        super(RoutineBase, self).__setattr__(key, value)

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to the routine instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        """
        if isinstance(value, (Param, Var, Constraint, Objective, ExpressionCalc)):
            value.om = self.om
            value.rtn = self
        if isinstance(value, Param):
            self.params[key] = value
            self.om.params[key] = None  # cp.Parameter
        if isinstance(value, Var):
            self.vars[key] = value
            self.om.vars[key] = None  # cp.Variable
        elif isinstance(value, Constraint):
            self.constrs[key] = value
            self.om.constrs[key] = None  # cp.Constraint
        elif isinstance(value, ExpressionCalc):
            self.exprs[key] = value
        elif isinstance(value, RParam):
            self.rparams[key] = value
        elif isinstance(value, RBaseService):
            self.services[key] = value

    def update(self, params=None, build_mats=True,):
        """
        Update the values of Parameters in the optimization model.

        This method is particularly important when some `RParams` are
        linked with system matrices.
        In such cases, setting `build_mats=True` is necessary to rebuild
        these matrices for the changes to take effect.
        This is common in scenarios involving topology changes, connection statuses,
        or load value modifications.
        If unsure, it is advisable to use `build_mats=True` as a precautionary measure.

        Parameters
        ----------
        params: Parameter, str, or list
            Parameter, Parameter name, or a list of parameter names to be updated.
            If None, all parameters will be updated.
        build_mats: bool
            True to rebuild the system matrices. Set to False to speed up the process
            if no system matrices are changed.
        """
        t0, _ = elapsed()
        re_finalize = False
        # sanitize input
        sparams = []
        if params is None:
            sparams = [val for val in self.params.values()]
            build_mats = True
        elif isinstance(params, Param):
            sparams = [params]
        elif isinstance(params, str):
            sparams = [self.params[params]]
        elif isinstance(params, list):
            sparams = [self.params[param] for param in params if isinstance(param, str)]
            for param in sparams:
                param.update()

        for param in sparams:
            if param.optz is None:  # means no_parse=True
                re_finalize = True
                break

        self.system.mats.build(force=build_mats)

        if re_finalize:
            logger.warning(f"<{self.class_name}> reinit OModel due to non-parametric change.")
            self.om.evaluate(force=True)
            self.om.finalize(force=True)

        results = self.om.update(params=sparams)
        t0, s0 = elapsed(t0)
        logger.debug(f"Update params in {s0}.")
        return results

    def __delattr__(self, name):
        """
        Overload the delattr function to unregister attributes.

        Parameters
        ----------
        name: str
            name of the attribute
        """
        self._unregister_attribute(name)
        if name == "obj":
            self.obj = None
        else:
            super().__delattr__(name)  # Call the superclass implementation

    def _unregister_attribute(self, name):
        """
        Unregister a pair of attributes from the routine instance.

        Called within ``__delattr__``, this is where the magic happens.
        Subclass attributes are automatically unregistered based on the variable type.
        """
        if name in self.vars:
            del self.vars[name]
            if name in self.om.vars:
                del self.om.vars[name]
        elif name in self.rparams:
            del self.rparams[name]
        elif name in self.constrs:
            del self.constrs[name]
            if name in self.om.constrs:
                del self.om.constrs[name]
        elif name in self.services:
            del self.services[name]

    def enable(self, name):
        """
        Enable a constraint by name.

        Parameters
        ----------
        name: str or list
            name of the constraint to be enabled
        """
        if isinstance(name, list):
            constr_act = []
            for n in name:
                if n not in self.constrs:
                    logger.warning(f"Constraint <{n}> not found.")
                    continue
                if not self.constrs[n].is_disabled:
                    logger.warning(f"Constraint <{n}> has already been enabled.")
                    continue
                self.constrs[n].is_disabled = False
                self.om.initialized = False
                constr_act.append(n)
            if len(constr_act) > 0:
                msg = ", ".join(constr_act)
                logger.warning(f"Turn on constraints: {msg}")
            return True

        if name in self.constrs:
            if not self.constrs[name].is_disabled:
                logger.warning(f"Constraint <{name}> has already been enabled.")
            else:
                self.constrs[name].is_disabled = False
                self.om.initialized = False
                logger.warning(f"Turn on constraint <{name}>.")
            return True

    def disable(self, name):
        """
        Disable a constraint by name.

        Parameters
        ----------
        name: str or list
            name of the constraint to be disabled
        """
        if isinstance(name, list):
            constr_act = []
            for n in name:
                if n not in self.constrs:
                    logger.warning(f"Constraint <{n}> not found.")
                elif self.constrs[n].is_disabled:
                    logger.warning(f"Constraint <{n}> has already been disabled.")
                else:
                    self.constrs[n].is_disabled = True
                    self.om.initialized = False
                    constr_act.append(n)
            if len(constr_act) > 0:
                msg = ", ".join(constr_act)
                logger.warning(f"Turn off constraints: {msg}")
            return True

        if name in self.constrs:
            if self.constrs[name].is_disabled:
                logger.warning(f"Constraint <{name}> has already been disabled.")
            else:
                self.constrs[name].is_disabled = True
                self.om.initialized = False
                logger.warning(f"Turn off constraint <{name}>.")
            return True

        logger.warning(f"Constraint <{name}> not found.")

    def _post_add_check(self):
        """
        Post-addition check.
        """
        # --- reset routine status ---
        self.initialized = False
        self.exec_time = 0.0
        self.exit_code = 0
        # --- reset symprocessor status ---
        self._syms = False
        # --- reset optimization model status ---
        self.om.initialized = False
        # --- reset OModel parser status ---
        self.om.parsed = False

    def addRParam(self,
                  name: str,
                  tex_name: Optional[str] = None,
                  info: Optional[str] = None,
                  src: Optional[str] = None,
                  unit: Optional[str] = None,
                  model: Optional[str] = None,
                  v: Optional[np.ndarray] = None,
                  indexer: Optional[str] = None,
                  imodel: Optional[str] = None,):
        """
        Add `RParam` to the routine.

        Parameters
        ----------
        name : str
            Name of this parameter. If not provided, `name` will be set
            to the attribute name.
        tex_name : str, optional
            LaTeX-formatted parameter name. If not provided, `tex_name`
            will be assigned the same as `name`.
        info : str, optional
            A description of this parameter
        src : str, optional
            Source name of the parameter.
        unit : str, optional
            Unit of the parameter.
        model : str, optional
            Name of the owner model or group.
        v : np.ndarray, optional
            External value of the parameter.
        indexer : str, optional
            Indexer of the parameter.
        imodel : str, optional
            Name of the owner model or group of the indexer.
        """
        item = RParam(name=name, tex_name=tex_name, info=info, src=src, unit=unit,
                      model=model, v=v, indexer=indexer, imodel=imodel)

        # add the parameter as an routine attribute
        setattr(self, name, item)

        # NOTE: manually register the owner of the parameter
        # This is skipped in ``addVars`` because of ``Var.__setattr__``
        item.rtn = self

        # check variable owner validity if given
        if model is not None:
            if item.model in self.system.groups.keys():
                item.is_group = True
                item.owner = self.system.groups[item.model]
            elif item.model in self.system.models.keys():
                item.owner = self.system.models[item.model]
            else:
                msg = f'Model indicator \'{item.model}\' of <{item.rtn.class_name}.{name}>'
                msg += ' is not a model or group. Likely a modeling error.'
                logger.warning(msg)

        self._post_add_check()
        return item

    def addService(self,
                   name: str,
                   value: np.ndarray,
                   tex_name: str = None,
                   unit: str = None,
                   info: str = None,
                   vtype: Type = None,
                   model: str = None,):
        """
        Add `ValueService` to the routine.

        Parameters
        ----------
        name : str
            Instance name.
        value : np.ndarray
            Value.
        tex_name : str, optional
            TeX name.
        unit : str, optional
            Unit.
        info : str, optional
            Description.
        vtype : Type, optional
            Variable type.
        model : str, optional
            Model name.
        """
        item = ValueService(name=name, tex_name=tex_name,
                            unit=unit, info=info,
                            vtype=vtype, value=value)
        # add the service as an routine attribute
        setattr(self, name, item)

        self._post_add_check()

        return item

    def addConstrs(self,
                   name: str,
                   e_str: str,
                   info: Optional[str] = None,
                   is_eq: Optional[str] = False,
                   ):
        """
        Add `Constraint` to the routine. to the routine.

        Parameters
        ----------
        name : str
            Constraint name. One should typically assigning the name directly because
            it will be automatically assigned by the model. The value of ``name``
            will be the symbol name to be used in expressions.
        e_str : str
            Constraint expression string.
        info : str, optional
            Descriptive information
        is_eq : str, optional
            Flag indicating if the constraint is an equality constraint. False indicates
            an inequality constraint in the form of `<= 0`.
        """
        item = Constraint(name=name, e_str=e_str, info=info, is_eq=is_eq)
        # add the constraint as an routine attribute
        setattr(self, name, item)

        self._post_add_check()

        return item

    def addVars(self,
                name: str,
                model: Optional[str] = None,
                shape: Optional[Union[int, tuple]] = None,
                tex_name: Optional[str] = None,
                info: Optional[str] = None,
                src: Optional[str] = None,
                unit: Optional[str] = None,
                horizon: Optional[RParam] = None,
                nonneg: Optional[bool] = False,
                nonpos: Optional[bool] = False,
                cplx: Optional[bool] = False,
                imag: Optional[bool] = False,
                symmetric: Optional[bool] = False,
                diag: Optional[bool] = False,
                psd: Optional[bool] = False,
                nsd: Optional[bool] = False,
                hermitian: Optional[bool] = False,
                boolean: Optional[bool] = False,
                integer: Optional[bool] = False,
                pos: Optional[bool] = False,
                neg: Optional[bool] = False,):
        """
        Add a variable to the routine.

        Parameters
        ----------
        name : str, optional
            Variable name. One should typically assigning the name directly because
            it will be automatically assigned by the model. The value of ``name``
            will be the symbol name to be used in expressions.
        model : str, optional
            Name of the owner model or group.
        shape : int or tuple, optional
            Shape of the variable. If is None, the shape of `model` will be used.
        info : str, optional
            Descriptive information
        unit : str, optional
            Unit
        tex_name : str
            LaTeX-formatted variable symbol. If is None, the value of `name` will be
            used.
        src : str, optional
            Source variable name. If is None, the value of `name` will be used.
        lb : str, optional
            Lower bound
        ub : str, optional
            Upper bound
        horizon : ams.routines.RParam, optional
            Horizon idx.
        nonneg : bool, optional
            Non-negative variable
        nonpos : bool, optional
            Non-positive variable
        cplx : bool, optional
            Complex variable
        imag : bool, optional
            Imaginary variable
        symmetric : bool, optional
            Symmetric variable
        diag : bool, optional
            Diagonal variable
        psd : bool, optional
            Positive semi-definite variable
        nsd : bool, optional
            Negative semi-definite variable
        hermitian : bool, optional
            Hermitian variable
        bool : bool, optional
            Boolean variable
        integer : bool, optional
            Integer variable
        pos : bool, optional
            Positive variable
        neg : bool, optional
            Negative variable
        """
        if model is None and shape is None:
            raise ValueError("Either model or shape must be specified.")
        item = Var(name=name, tex_name=tex_name,
                   info=info, src=src, unit=unit,
                   model=model, shape=shape, horizon=horizon,
                   nonneg=nonneg, nonpos=nonpos,
                   cplx=cplx, imag=imag,
                   symmetric=symmetric, diag=diag,
                   psd=psd, nsd=nsd, hermitian=hermitian,
                   boolean=boolean, integer=integer,
                   pos=pos, neg=neg, )

        # add the variable as an routine attribute
        setattr(self, name, item)

        # check variable owner validity if given
        if model is not None:
            if item.model in self.system.groups.keys():
                item.is_group = True
                item.owner = self.system.groups[item.model]
            elif item.model in self.system.models.keys():
                item.owner = self.system.models[item.model]
            else:
                msg = (
                    f"Model indicator '{item.model}' of <{item.rtn.class_name}.{name}>"
                )
                msg += " is not a model or group. Likely a modeling error."
                logger.warning(msg)

        self._post_add_check()

        return item

    def _initial_guess(self):
        """
        Generate initial guess for the optimization model.
        """
        raise NotImplementedError



"""
Benchmark functions.
"""

import datetime
import sys
import importlib.metadata as importlib_metadata
import logging

try:
    import pandapower as pdp
    PANDAPOWER_AVAILABLE = True
except ImportError:
    PANDAPOWER_AVAILABLE = False
    logging.warning("pandapower is not available. Some functionalities will be disabled.")

from andes.utils.misc import elapsed

import ams

logger = logging.getLogger(__name__)

_failed_time = -1
_failed_obj = -1

cols_pre = ['ams_mats', 'ams_parse', 'ams_eval', 'ams_final', 'ams_postinit']


def get_tool_versions(tools=None):
    """
    Get the current time, Python version, and versions of specified tools.

    Parameters
    ----------
    tools : list of str, optional
        List of tool names to check versions. If None, a default list of tools is used.

    Returns
    -------
    dict
        A dictionary containing the tool names and their versions.
    """
    if tools is None:
        tools = ['ltbams', 'andes', 'cvxpy',
                 'gurobipy', 'mosek', 'piqp',
                 'pandapower', 'numba']

    # Get current time and Python version
    last_run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    python_version = sys.version

    # Collect tool versions
    tool_versions = {}
    for tool in tools:
        try:
            version = importlib_metadata.version(tool)
            tool_versions[tool] = version
        except importlib_metadata.PackageNotFoundError:
            logger.error(f"Package {tool} not found.")
            tool_versions[tool] = "Not installed"

    # Print the results in a formatted way
    logger.warning(f"Last run time: {last_run_time}")
    logger.warning(f"Python: {python_version}\n")

    # Calculate the width of the columns
    max_tool_length = max(len(tool) for tool in tool_versions)
    max_version_length = max(len(version) for version in tool_versions.values())

    # Print the header
    logger.warning(f"{'Tool':<{max_tool_length}}  {'Version':<{max_version_length}}")
    logger.warning(f"{'-' * max_tool_length}  {'-' * max_version_length}")

    # Print each tool and its version
    for tool, version in tool_versions.items():
        logger.warning(f"{tool:<{max_tool_length}}  {version:<{max_version_length}}")

    return tool_versions


def time_routine_solve(system, routine='DCOPF', **kwargs):
    """
    Run the specified routine with the given solver and method.

    Parameters
    ----------
    system : ams.System
        The system object containing the routine.
    routine : str, optional
        The name of the routine to run. Defaults to 'DCOPF'.

    Other Parameters
    ----------------
    solver : str, optional
        The solver to use.
    ignore_dpp : bool, optional
        Whether to ignore DPP. Defaults to True.

    Returns
    -------
    tuple
        A tuple containing the elapsed time (s) and the objective value ($).
    """
    rtn = system.routines[routine]
    solver = kwargs.get('solver', None)
    try:
        t, _ = elapsed()
        rtn.run(**kwargs)
        _, s0 = elapsed(t)
        elapsed_time = float(s0.split(' ')[0])
        obj_value = rtn.obj.v
    except Exception as e:
        logger.error(f"Error running routine {routine} with solver {solver}: {e}")
        elapsed_time = _failed_time
        obj_value = _failed_obj
    return elapsed_time, obj_value


def pre_solve(system, routine):
    """
    Time the routine preparation process.

    Parameters
    ----------
    system : ams.System
        The system object containing the routine.
    routine : str
        The name of the routine to prepare

    Returns
    -------
    dict
        A dictionary containing the preparation times in seconds for each step:
        'mats', 'parse', 'evaluate', 'finalize', 'postinit'.
    """
    rtn = system.routines[routine]

    # Initialize AMS
    # --- matrices build ---
    t_mats, _ = elapsed()
    system.mats.build(force=True)
    _, s_mats = elapsed(t_mats)

    # --- code generation ---
    t_parse, _ = elapsed()
    rtn.om.parse(force=True)
    _, s_parse = elapsed(t_parse)

    # --- code evaluation ---
    t_evaluate, _ = elapsed()
    rtn.om.evaluate(force=True)
    _, s_evaluate = elapsed(t_evaluate)

    # --- problem finalization ---
    t_finalize, _ = elapsed()
    rtn.om.finalize(force=True)
    _, s_finalize = elapsed(t_finalize)

    # --- rest init process ---
    t_postinit, _ = elapsed()
    rtn.init()
    _, s_postinit = elapsed(t_postinit)

    s_float = [float(s.split(' ')[0]) for s in [s_mats, s_parse, s_evaluate, s_finalize, s_postinit]]

    pre_time = dict(zip(cols_pre, s_float))
    return pre_time


def time_pdp_dcopf(ppn):
    """
    Test the execution time of DCOPF using pandapower.

    Parameters
    ----------
    ppn : pandapowerNet
        The pandapower network object.

    Returns
    -------
    tuple
        A tuple containing the elapsed time (s) and the objective value ($).
    """
    try:
        t_pdp, _ = elapsed()
        pdp.rundcopp(ppn)
        _, s_pdp = elapsed(t_pdp)
        elapsed_time = float(s_pdp.split(' ')[0])
        obj_value = ppn.res_cost
    except Exception as e:
        logger.error(f"Error running pandapower: {e}")
        elapsed_time = _failed_time
        obj_value = _failed_obj
    return elapsed_time, obj_value


def time_routine(system, routine='DCOPF', solvers=['CLARABEL'],
                 **kwargs):
    """
    Time the specified routine with the given solvers.

    Parameters
    ----------
    system : ams.System
        The system object containing the routine.
    routine : str, optional
        The name of the routine to run. Defaults to 'DCOPF'.
    solvers : list of str, optional
        List of solvers to use. Defaults to ['CLARABEL'].

    Other Parameters
    ----------------
    ignore_dpp : bool, optional
        Whether to ignore DPP. Defaults to True.

    Returns
    -------
    tuple
        A tuple containing the preparation times and the solution times in
        seconds for each solver.
    """
    pre_time = pre_solve(system, routine)
    sol = {f'{solver}': {'time': 0, 'obj': 0} for solver in solvers}

    for solver in solvers:
        if solver != 'pandapower':
            s, obj = time_routine_solve(system, routine, solver=solver, **kwargs)
            sol[solver]['time'] = s
            sol[solver]['obj'] = obj
        elif solver == 'pandapower' and PANDAPOWER_AVAILABLE and routine == 'DCOPF':
            ppc = ams.io.pypower.system2ppc(system)
            ppn = pdp.converter.from_ppc(ppc, f_hz=system.config.freq)
            s, obj = time_pdp_dcopf(ppn)
            sol[solver]['time'] = s
            sol[solver]['obj'] = obj
        else:
            sol[solver]['time'] = _failed_time
            sol[solver]['obj'] = _failed_obj

    return pre_time, sol


def time_dcopf_with_lf(system, solvers=['CLARABEL'], load_factors=[1], ignore_dpp=False):
    """
    Time the execution of DCOPF with varying load factors.

    Parameters
    ----------
    system : ams.System
        The system object containing the routine.
    solvers : list of str, optional
        List of solvers to use. Defaults to ['CLARABEL'].
    load_factors : list of float, optional
        List of load factors to apply. Defaults to None.
    ignore_dpp : bool, optional
        Whether to ignore DPP.

    Returns
    -------
    tuple
        A tuple containing the list of times and the list of objective values.
    """
    pre_time = pre_solve(system, 'DCOPF')
    sol = {f'{solver}': {'time': 0, 'obj': 0} for solver in solvers}

    pd0 = system.PQ.p0.v.copy()
    pq_idx = system.PQ.idx.v

    for solver in solvers:
        if solver != 'pandapower':
            obj_all = 0
            t_all, _ = elapsed()
            for lf_k in load_factors:
                system.PQ.set(src='p0', attr='v', idx=pq_idx, value=lf_k * pd0)
                system.DCOPF.update(params=['pd'])
                _, obj = time_routine_solve(system, 'DCOPF',
                                            solver=solver, ignore_dpp=ignore_dpp)
                obj_all += obj
            _, s_all = elapsed(t_all)
            system.PQ.set(src='p0', attr='v', idx=pq_idx, value=pd0)
            s = float(s_all.split(' ')[0])
            sol[solver]['time'] = s
            sol[solver]['obj'] = obj_all
        elif solver == 'pandapower' and PANDAPOWER_AVAILABLE:
            ppc = ams.io.pypower.system2ppc(system)
            ppn = pdp.converter.from_ppc(ppc, f_hz=system.config.freq)
            p_mw0 = ppn.load['p_mw'].copy()
            t_all, _ = elapsed()
            obj_all = 0
            for lf_k in load_factors:
                ppn.load['p_mw'] = lf_k * p_mw0
                _, obj = time_pdp_dcopf(ppn)
                obj_all += obj
            _, s_all = elapsed(t_all)
            s = float(s_all.split(' ')[0])
            sol[solver]['time'] = s
            sol[solver]['obj'] = obj_all
        else:
            sol[solver]['time'] = _failed_time
            sol[solver]['obj'] = _failed_obj

    return pre_time, sol



"""
Test module MatProcessor.
"""

import unittest
import os

import numpy as np

import ams
from ams.core.matprocessor import MatProcessor, MParam
from ams.shared import sps


class TestMatProcessorBasic(unittest.TestCase):
    """
    Test basic functionality of MatProcessor.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("matpower/case300.m"),
                           default_config=True, no_output=True)
        self.nR = self.ss.Region.n
        self.nb = self.ss.Bus.n
        self.nl = self.ss.Line.n
        self.ng = self.ss.StaticGen.n
        self.nsh = self.ss.Shunt.n
        self.nD = self.ss.StaticLoad.n

        self.mats = MatProcessor(self.ss)
        self.mats.build()

    def test_MParams_owner(self):
        """
        Tesst MParams owner before system initialization.
        """
        self.assertIs(self.mats.Cft.owner, self.mats)
        self.assertIs(self.mats.CftT.owner, self.mats)
        self.assertIs(self.mats.Cg.owner, self.mats)
        self.assertIs(self.mats.Cl.owner, self.mats)
        self.assertIs(self.mats.Csh.owner, self.mats)
        self.assertIs(self.mats.Bbus.owner, self.mats)
        self.assertIs(self.mats.Bf.owner, self.mats)
        self.assertIs(self.mats.Pbusinj.owner, self.mats)
        self.assertIs(self.mats.Pfinj.owner, self.mats)
        self.assertIs(self.mats.PTDF.owner, self.mats)
        self.assertIs(self.mats.LODF.owner, self.mats)

    def test_MParam_export_csv(self):
        """
        Test MParams export.
        """
        # --- path is not given ---
        exported_csv = self.mats.Cft.export_csv()
        self.assertTrue(os.path.exists(exported_csv))
        os.remove(exported_csv)

        # --- path is given ---
        path = 'CASE300_Cft.csv'
        exported_csv = self.mats.Cft.export_csv(path)
        self.assertTrue(os.path.exists(path))
        os.remove(path)

    def test_MParams_export(self) -> None:
        """
        Test MParams export.
        """
        cft = self.mats.Cft.export_csv()
        self.assertTrue(os.path.exists(cft))
        os.remove(cft)

        cftt = self.mats.CftT.export_csv()
        self.assertTrue(os.path.exists(cftt))
        os.remove(cftt)

        cg = self.mats.Cg.export_csv()
        self.assertTrue(os.path.exists(cg))
        os.remove(cg)

        cl = self.mats.Cl.export_csv()
        self.assertTrue(os.path.exists(cl))
        os.remove(cl)

        csh = self.mats.Csh.export_csv()
        self.assertTrue(os.path.exists(csh))
        os.remove(csh)

        bbus = self.mats.Bbus.export_csv()
        self.assertTrue(os.path.exists(bbus))
        os.remove(bbus)

        bf = self.mats.Bf.export_csv()
        self.assertTrue(os.path.exists(bf))
        os.remove(bf)

        ptdf = self.mats.PTDF.export_csv()
        self.assertTrue(os.path.exists(ptdf))
        os.remove(ptdf)

        lodf = self.mats.LODF.export_csv()
        self.assertTrue(os.path.exists(lodf))
        os.remove(lodf)

    def test_MParam_instance(self):
        """
        Test `MParam` instantiate.
        """
        one_vec = MParam(v=sps.csr_matrix(np.ones(self.ss.Bus.n)))
        # check if `_v` is `sps.csr_matrix` instance
        self.assertIsInstance(one_vec._v, sps.csr_matrix)
        # check if `v` is 1D-array
        self.assertEqual(one_vec.v.shape, (self.ss.Bus.n,))

    def test_c(self):
        """
        Test connectivity matrices.
        """
        # Test `Cg`
        self.assertIsInstance(self.mats.Cg._v, (sps.csr_matrix, sps.lil_matrix))
        self.assertIsInstance(self.mats.Cg.v, np.ndarray)
        self.assertEqual(self.mats.Cg._v.max(), 1)
        np.testing.assert_equal(self.mats.Cg._v.sum(axis=0), np.ones((1, self.ng)))

        # Test `Cl`
        self.assertIsInstance(self.mats.Cl._v, (sps.csr_matrix, sps.lil_matrix))
        self.assertIsInstance(self.mats.Cl.v, np.ndarray)
        self.assertEqual(self.mats.Cl._v.max(), 1)
        np.testing.assert_equal(self.mats.Cl._v.sum(axis=0), np.ones((1, self.nD)))

        # Test `Csh`
        self.assertIsInstance(self.mats.Csh._v, (sps.csr_matrix, sps.lil_matrix))
        self.assertIsInstance(self.mats.Csh.v, np.ndarray)
        self.assertEqual(self.mats.Csh._v.max(), 1)
        np.testing.assert_equal(self.mats.Csh._v.sum(axis=0), np.ones((1, self.nsh)))

        # Test `Cft`
        self.assertIsInstance(self.mats.Cft._v, (sps.csr_matrix, sps.lil_matrix))
        self.assertIsInstance(self.mats.Cft.v, np.ndarray)
        self.assertEqual(self.mats.Cft._v.max(), 1)
        np.testing.assert_equal(self.mats.Cft._v.sum(axis=0), np.zeros((1, self.nl)))

    def test_b_calc(self):
        """
        Test `b_calc`.
        """
        b = self.mats._calc_b()
        self.assertIsInstance(b, np.ndarray)
        self.assertEqual(b.shape, (self.nl,))

    def test_bf(self):
        """
        Test `Bf`.
        """
        self.assertIsInstance(self.mats.Bf._v, (sps.csr_matrix, sps.lil_matrix))
        np.testing.assert_equal(self.mats.Bf._v.shape, (self.nl, self.nb))

    def test_bbus(self):
        """
        Test `Bbus`.
        """
        self.assertIsInstance(self.mats.Bbus._v, (sps.csr_matrix, sps.lil_matrix))
        np.testing.assert_equal(self.mats.Bbus._v.shape, (self.nb, self.nb))

    def test_pfinj(self):
        """
        Test `Pfinj`.
        """
        self.assertIsInstance(self.mats.Pfinj._v, np.ndarray)
        np.testing.assert_equal(self.mats.Pfinj._v.shape, (self.nl,))

    def test_pbusinj(self):
        """
        Test `Pbusinj`.
        """
        self.assertIsInstance(self.mats.Pbusinj._v, np.ndarray)
        np.testing.assert_equal(self.mats.Pbusinj._v.shape, (self.nb,))


class TestBuildPTDF(unittest.TestCase):
    """
    Test build PTDF.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case('matpower/case14.m'),
                           setup=True, default_config=True, no_output=True)
        self.nl = self.ss.Line.n
        self.nb = self.ss.Bus.n

        self.assertFalse(self.ss.mats.initialized)
        # NOTE: here we test `no_store=True` option
        self.ptdf_full = self.ss.mats.build_ptdf(no_store=True)
        self.assertTrue(self.ss.mats.initialized)

    def test_ptdf_before_mat_init(self):
        """
        Test PTDF before MatProcessor initialization `mats.init()`.
        """
        self.assertIsNone(self.ss.mats.PTDF._v)
        _ = self.ss.mats.build_ptdf(no_store=False)
        self.assertEqual(self.ss.mats.PTDF._v.shape, (self.nl, self.nb))

    def test_ptdf_lines(self):
        """
        Test PTDF with `line` inputs.
        """
        # NOTE: when `line` is given, `PTDF` should not be stored even with `no_store=False`

        # input str
        ptdf_l2 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2], no_store=False)
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_l2, self.ptdf_full[2, :])

        # input list with single element
        ptdf_l2p = self.ss.mats.build_ptdf(line=[self.ss.Line.idx.v[2]], no_store=False)
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_l2p, self.ptdf_full[[2], :])

        # input list with multiple elements
        ptdf_l23 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2:4], no_store=False)
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_l23, self.ptdf_full[2:4, :])

    def test_ptdf_incremental_lines(self):
        """
        Test PTDF incremental build with `line` inputs.
        """
        # input str
        ptdf_l2 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2], incremental=True, no_store=False,
                                          no_tqdm=True)
        np.testing.assert_array_almost_equal(ptdf_l2.todense(),
                                             self.ptdf_full[[2], :])
        self.assertTrue(sps.isspmatrix_lil(ptdf_l2))
        self.assertIsNone(self.ss.mats.PTDF._v)

        # input list with single element
        ptdf_l2p = self.ss.mats.build_ptdf(line=[self.ss.Line.idx.v[2]], incremental=True, no_store=False,
                                           no_tqdm=False)
        np.testing.assert_array_almost_equal(ptdf_l2p.todense(),
                                             self.ptdf_full[[2], :])
        self.assertTrue(sps.isspmatrix_lil(ptdf_l2p))
        self.assertIsNone(self.ss.mats.PTDF._v)

        # input list with multiple elements
        ptdf_l23 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2:4], incremental=True, no_store=False)
        np.testing.assert_array_almost_equal(ptdf_l23.todense(),
                                             self.ptdf_full[2:4, :])

    def test_ptdf_incremental_step(self):
        """
        Test PTDF incremental build with step.
        """
        # step < line length
        ptdf_c1 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=1)
        self.assertTrue(sps.isspmatrix_lil(ptdf_c1))
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_c1.todense(),
                                             self.ptdf_full[2:4, :],)

        # step = line length
        ptdf_c2 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=2)
        self.assertTrue(sps.isspmatrix_lil(ptdf_c2))
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_c2.todense(),
                                             self.ptdf_full[2:4, :],)

        # step > line length
        ptdf_c5 = self.ss.mats.build_ptdf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=5)
        self.assertTrue(sps.isspmatrix_lil(ptdf_c5))
        self.assertIsNone(self.ss.mats.PTDF._v)
        np.testing.assert_array_almost_equal(ptdf_c5.todense(),
                                             self.ptdf_full[2:4, :],)


class TestBuildLODF(unittest.TestCase):
    """
    Test build LODF.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case('matpower/case14.m'),
                           setup=True, default_config=True, no_output=True)
        self.nl = self.ss.Line.n
        self.nb = self.ss.Bus.n
        self.dec = 4

        self.assertFalse(self.ss.mats.initialized)
        self.lodf_full = self.ss.mats.build_lodf(no_store=True)
        self.assertTrue(self.ss.mats.initialized)

    def test_lodf_before_ptdf(self):
        """
        Test LODF before PTDF.
        """
        self.assertIsNone(self.ss.mats.PTDF._v)
        self.assertIsNone(self.ss.mats.LODF._v)
        _ = self.ss.mats.build_lodf(no_store=False)
        self.assertEqual(self.ss.mats.LODF._v.shape, (self.nl, self.nl))
        # PTDF should not be stored for requested building
        self.assertIsNone(self.ss.mats.PTDF._v)

    def test_lodf_lines(self):
        """
        Test LODF with `line` inputs.
        """
        # input str
        lodf_l2 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2], no_store=False)
        np.testing.assert_array_almost_equal(lodf_l2, self.lodf_full[:, 2], decimal=self.dec)

        # input list with single element
        lodf_l2p = self.ss.mats.build_lodf(line=[self.ss.Line.idx.v[2]], no_store=False)
        np.testing.assert_array_almost_equal(lodf_l2p, self.lodf_full[:, [2]], decimal=self.dec)

        # input list with multiple elements
        lodf_l23 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2:4], no_store=False)
        np.testing.assert_array_almost_equal(lodf_l23, self.lodf_full[:, 2:4], decimal=self.dec)

    def test_lodf_incremental_lines(self):
        """
        Test LODF incremental build with `line` inputs.
        """
        # input str
        lodf_l2 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2], incremental=True, no_store=False)
        np.testing.assert_array_almost_equal(lodf_l2.todense(),
                                             self.lodf_full[:, [2]],
                                             decimal=self.dec)
        self.assertTrue(sps.isspmatrix_lil(lodf_l2))

        # input list with single element
        lodf_l2p = self.ss.mats.build_lodf(line=[self.ss.Line.idx.v[2]], incremental=True, no_store=False)
        np.testing.assert_array_almost_equal(lodf_l2p.todense(),
                                             self.lodf_full[:, [2]],
                                             decimal=self.dec)
        self.assertTrue(sps.isspmatrix_lil(lodf_l2p))

        # input list with multiple elements
        lodf_l23 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2:4], incremental=True, no_store=False)
        np.testing.assert_array_almost_equal(lodf_l23.todense(),
                                             self.lodf_full[:, 2:4],
                                             decimal=self.dec)

    def test_lodf_incremental_step(self):
        """
        Test LODF incremental build with step.
        """
        # step < line length
        lodf_c1 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=1)
        self.assertTrue(sps.isspmatrix_lil(lodf_c1))
        np.testing.assert_array_almost_equal(lodf_c1.todense(),
                                             self.lodf_full[:, 2:4],
                                             decimal=self.dec)

        # step = line length
        lodf_c2 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=2)
        self.assertTrue(sps.isspmatrix_lil(lodf_c2))
        np.testing.assert_array_almost_equal(lodf_c2.todense(),
                                             self.lodf_full[:, 2:4],
                                             decimal=self.dec)

        # step > line length
        lodf_c5 = self.ss.mats.build_lodf(line=self.ss.Line.idx.v[2:4], incremental=True,
                                          no_store=False, step=5)
        self.assertTrue(sps.isspmatrix_lil(lodf_c5))
        np.testing.assert_array_almost_equal(lodf_c5.todense(),
                                             self.lodf_full[:, 2:4],
                                             decimal=self.dec)


class TestBuildOTDF(unittest.TestCase):
    """
    Test build OTDF.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case('matpower/case14.m'),
                           setup=True, default_config=True, no_output=True)
        self.nl = self.ss.Line.n
        self.nb = self.ss.Bus.n
        self.dec = 4
        _ = self.ss.mats.build()

    def test_otdf_dense_build(self):
        _ = self.ss.mats.build_ptdf(no_store=False, incremental=False)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=False)
        self.otdf_full = self.ss.mats.build_otdf()
        self.assertEqual(self.otdf_full.shape, (self.nl, self.nb))

    def test_otdf_sparse_build(self):
        # --- both PTDF and LODF are dense ---
        _ = self.ss.mats.build_ptdf(no_store=False, incremental=False)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=False)
        otdf_dense = self.ss.mats.build_otdf()

        _ = self.ss.mats.build_ptdf(no_store=False, incremental=True)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=True)
        otdf_sparse = self.ss.mats.build_otdf()
        np.testing.assert_array_almost_equal(otdf_sparse.todense(), otdf_dense,
                                             decimal=self.dec)

        # --- PTDF is dense and LODF is sparse ---
        _ = self.ss.mats.build_ptdf(no_store=False, incremental=False)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=True)
        otdf_sps1 = self.ss.mats.build_otdf()
        np.testing.assert_array_almost_equal(otdf_sps1, otdf_dense,
                                             decimal=self.dec)

        # --- PTDF is sparse and LODF is dense ---
        _ = self.ss.mats.build_ptdf(no_store=False, incremental=True)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=False)
        otdf_sps2 = self.ss.mats.build_otdf()
        np.testing.assert_array_almost_equal(otdf_sps2.todense(), otdf_dense,
                                             decimal=self.dec)

    def test_otdf_lines(self):
        """
        Test OTDF with `line` inputs.
        """
        # --- both PTDF and LODF are dense ---
        _ = self.ss.mats.build_ptdf(no_store=False, incremental=False)
        _ = self.ss.mats.build_lodf(no_store=False, incremental=False)

        # input str
        otdf_l2 = self.ss.mats.build_otdf(line=self.ss.Line.idx.v[2])
        self.assertEqual(otdf_l2.shape, (self.nl, self.nb))

        # input list with single element
        otdf_l2p = self.ss.mats.build_otdf(line=[self.ss.Line.idx.v[2]])
        self.assertEqual(otdf_l2p.shape, (self.nl, self.nb))

        # input list with multiple elements
        otdf_l23 = self.ss.mats.build_otdf(line=self.ss.Line.idx.v[2:4])
        self.assertEqual(otdf_l23.shape, (self.nl, self.nb))



"""
PYPOWER version.
"""


def ppver(*args):
    """
    Returns PYPOWER version info for current installation.

    @author: Ray Zimmerman (PSERC Cornell)
    """

    ver = {'Name': 'PYPOWER',
           'Version': '5.1.4',
           'Release':  '',
           'Date': '27-June-2018'}

    return ver



import unittest

import ams
import ams.benchmarks as bp


def are_packages_available(*packages):
    """
    Check if the specified packages are available.

    Parameters
    ----------
    *packages : str
        Names of the packages to check.

    Returns
    -------
    bool
        True if all packages are available, False otherwise.
    """
    for package in packages:
        try:
            __import__(package)
        except ImportError:
            return False
    return True


def require_packages(*packages):
    """
    Decorator to skip a test if the specified packages are not available.

    Parameters
    ----------
    *packages : str
        Names of the packages required for the test.

    Returns
    -------
    function
        The decorated test function.
    """
    def decorator(test_func):
        return unittest.skipIf(
            not are_packages_available(*packages),
            f"Skipping test because one or more required packages are not available: {', '.join(packages)}"
        )(test_func)
    return decorator


class TestBenchmarks(unittest.TestCase):
    """
    Test module for benchmarks.py.
    """

    def test_get_tool_versions(self):
        self.assertIsInstance(bp.get_tool_versions(), dict)

    def test_pre_solve_rted(self):
        ss = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True, default_config=True, no_output=True)
        pre_time = bp.pre_solve(ss, 'RTED')
        self.assertIsInstance(pre_time, dict)
        self.assertEqual(len(pre_time), len(bp.cols_pre))
        for v in pre_time.values():
            self.assertIsInstance(v, float)

    def test_dcopf_solve(self):
        ss = ams.load(ams.get_case('matpower/case5.m'),
                      setup=True, default_config=True, no_output=True)
        s, obj = bp.time_routine_solve(ss, routine='DCOPF', solver='CLARABEL', ignore_dpp=False)
        self.assertGreater(s, 0)
        self.assertGreater(obj, 0)

    def test_pre_solve_ed(self):
        ss = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True, default_config=True, no_output=True)
        pre_time = bp.pre_solve(ss, 'ED')
        self.assertIsInstance(pre_time, dict)
        self.assertEqual(len(pre_time), len(bp.cols_pre))
        for v in pre_time.values():
            self.assertIsInstance(v, float)

    def test_ed_solve(self):
        ss = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True, default_config=True, no_output=True)
        s, obj = bp.time_routine_solve(ss, 'ED', solver='CLARABEL', ignore_dpp=True)
        self.assertGreater(s, 0)
        self.assertGreater(obj, 0)

    @require_packages('pandapower')
    def time_pdp_dcopf(self):
        ss = ams.load(ams.get_case('matpower/case5.m'),
                      setup=True, default_config=True, no_output=True)
        ppc = ams.io.pypower.system2ppc(ss)
        ppn = ams.io.pandapower.converter.from_ppc(ppc, f_hz=ss.config.freq)
        s, obj = bp.time_pdp_dcopf(ppn)
        self.assertGreater(s, 0)
        self.assertGreater(obj, 0)

    def test_time_dcopf(self):
        ss = ams.load(ams.get_case('matpower/case5.m'),
                      setup=True, default_config=True, no_output=True)
        pre_time, sol = bp.time_routine(ss,
                                        routine='DCOPF',
                                        solvers=['CLARABEL', 'SCIP', 'pandapower'], ignore_dpp=False)
        for v in pre_time.values():
            self.assertGreaterEqual(v, 0)

        self.assertGreater(sol['CLARABEL']['time'], 0)
        self.assertGreater(sol['SCIP']['time'], 0)
        self.assertAlmostEqual(sol['CLARABEL']['obj'],
                               sol['SCIP']['obj'],
                               places=2)
        if not are_packages_available('pandapower'):
            self.assertEqual(sol['pandapower']['time'], -1)
            self.assertEqual(sol['pandapower']['obj'], -1)
        else:
            self.assertGreater(sol['pandapower']['obj'], 0)
            self.assertAlmostEqual(sol['CLARABEL']['obj'],
                                   sol['pandapower']['obj'],
                                   places=2)

    def test_time_rted(self):
        ss = ams.load(ams.get_case('5bus/pjm5bus_demo.xlsx'),
                      setup=True, default_config=True, no_output=True)
        pre_time, sol = bp.time_routine(ss,
                                        routine='RTED',
                                        solvers=['CLARABEL', 'SCIP', 'pandapower'], ignore_dpp=False)
        for v in pre_time.values():
            self.assertGreaterEqual(v, 0)

        self.assertGreater(sol['CLARABEL']['time'], 0)
        self.assertGreater(sol['SCIP']['time'], 0)
        self.assertAlmostEqual(sol['CLARABEL']['obj'],
                               sol['SCIP']['obj'],
                               places=2)
        self.assertEqual(sol['pandapower']['time'], -1)
        self.assertEqual(sol['pandapower']['obj'], -1)

    def test_time_dcopf_with_lf(self):
        ss = ams.load(ams.get_case('matpower/case5.m'),
                      setup=True, default_config=True, no_output=True)
        pre_time, sol = bp.time_dcopf_with_lf(ss, solvers=['CLARABEL', 'SCIP'], load_factors=[
                                              1, 0.5, 0.25], ignore_dpp=False)
        self.assertEqual(len(pre_time), len(bp.cols_pre))
        self.assertAlmostEqual(sol['CLARABEL']['obj'],
                               sol['SCIP']['obj'],
                               places=2)
